# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for Computer entities"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import logging
import os
import six

from aiida import transports, schedulers
from aiida.common import exceptions
from aiida.manage.manager import get_manager
from aiida.plugins import SchedulerFactory, TransportFactory

from . import entities
from . import users

__all__ = ('Computer',)


class Computer(entities.Entity):
    """
    Base class to map a node in the DB + its permanent repository counterpart.

    Stores attributes starting with an underscore.

    Caches files and attributes before the first save, and saves everything only on store().
    After the call to store(), attributes cannot be changed.

    Only after storing (or upon loading from uuid) metadata can be modified
    and in this case they are directly set on the db.

    In the plugin, also set the _plugin_type_string, to be set in the DB in the 'type' field.
    """
    # pylint: disable=too-many-public-methods

    _logger = logging.getLogger(__name__)

    PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL = 'minimum_scheduler_poll_interval'  # pylint: disable=invalid-name
    PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL__DEFAULT = 10.  # pylint: disable=invalid-name
    PROPERTY_WORKDIR = 'workdir'
    PROPERTY_SHEBANG = 'shebang'

    class Collection(entities.Collection):
        """The collection of Computer entries."""

        def list_names(self):
            """Return a list with all the names of the computers in the DB."""
            return self._backend.computers.list_names()

        def delete(self, id):  # pylint: disable=redefined-builtin,invalid-name
            """Delete the computer with the given id"""
            return self._backend.computers.delete(id)

    def __init__(self, name, hostname, description='', transport_type='', scheduler_type='', workdir=None,
                 backend=None):
        """Construct a new computer"""
        # pylint: disable=too-many-arguments
        backend = backend or get_manager().get_backend()
        model = backend.computers.create(
            name=name,
            hostname=hostname,
            description=description,
            transport_type=transport_type,
            scheduler_type=scheduler_type)
        super(Computer, self).__init__(model)
        if workdir is not None:
            self.set_workdir(workdir)

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, str(self))

    def __str__(self):
        return '{} ({}), pk: {}'.format(self.name, self.hostname, self.pk)

    @property
    def full_text_info(self):
        """
        Return a (multiline) string with a human-readable detailed information on this computer.

        :rypte: str
        """
        ret_lines = []
        ret_lines.append("Computer name:     {}".format(self.name))
        ret_lines.append(" * PK:             {}".format(self.pk))
        ret_lines.append(" * UUID:           {}".format(self.uuid))
        ret_lines.append(" * Description:    {}".format(self.description))
        ret_lines.append(" * Hostname:       {}".format(self.hostname))
        ret_lines.append(" * Transport type: {}".format(self.get_transport_type()))
        ret_lines.append(" * Scheduler type: {}".format(self.get_scheduler_type()))
        ret_lines.append(" * Work directory: {}".format(self.get_workdir()))
        ret_lines.append(" * Shebang:        {}".format(self.get_shebang()))
        ret_lines.append(" * mpirun command: {}".format(" ".join(self.get_mpirun_command())))
        def_cpus_machine = self.get_default_mpiprocs_per_machine()
        if def_cpus_machine is not None:
            ret_lines.append(" * Default number of cpus per machine: {}".format(def_cpus_machine))
        # pylint: disable=fixme
        # TODO: Put back following line when we port Node to new backend system
        # ret_lines.append(" * Used by:        {} nodes".format(len(self._dbcomputer.dbnodes.all())))

        ret_lines.append(" * prepend text:")
        if self.get_prepend_text().strip():
            for line in self.get_prepend_text().split('\n'):
                ret_lines.append("   {}".format(line))
        else:
            ret_lines.append("   # No prepend text.")
        ret_lines.append(" * append text:")
        if self.get_append_text().strip():
            for line in self.get_append_text().split('\n'):
                ret_lines.append("   {}".format(line))
        else:
            ret_lines.append("   # No append text.")

        return "\n".join(ret_lines)

    @property
    def logger(self):
        return self._logger

    # region validation

    @classmethod
    def _name_validator(cls, name):
        """
        Validates the name.
        """
        if not name.strip():
            raise exceptions.ValidationError("No name specified")

    @classmethod
    def _hostname_validator(cls, hostname):
        """
        Validates the hostname.
        """
        if not hostname.strip():
            raise exceptions.ValidationError("No hostname specified")

    @classmethod
    def _description_validator(cls, description):
        """
        Validates the description.
        """
        # The description is always valid

    @classmethod
    def _transport_type_validator(cls, transport_type):
        """
        Validates the transport string.
        """
        if transport_type not in transports.Transport.get_valid_transports():
            raise exceptions.ValidationError("The specified transport is not a valid one")

    @classmethod
    def _scheduler_type_validator(cls, scheduler_type):
        """
        Validates the transport string.
        """
        if scheduler_type not in schedulers.Scheduler.get_valid_schedulers():
            raise exceptions.ValidationError("The specified scheduler is not a valid one")

    @classmethod
    def _prepend_text_validator(cls, prepend_text):
        """
        Validates the prepend text string.
        """
        # no validation done

    @classmethod
    def _append_text_validator(cls, append_text):
        """
        Validates the append text string.
        """
        # no validation done

    @classmethod
    def _workdir_validator(cls, workdir):
        """
        Validates the transport string.
        """
        if not workdir.strip():
            raise exceptions.ValidationError("No workdir specified")

        try:
            convertedwd = workdir.format(username="test")
        except KeyError as exc:
            raise exceptions.ValidationError("In workdir there is an unknown replacement field {}".format(exc.args[0]))
        except ValueError as exc:
            raise exceptions.ValidationError("Error in the string: '{}'".format(exc))

        if not os.path.isabs(convertedwd):
            raise exceptions.ValidationError("The workdir must be an absolute path")

    def _mpirun_command_validator(self, mpirun_cmd):
        """
        Validates the mpirun_command variable. MUST be called after properly
        checking for a valid scheduler.
        """
        if not isinstance(mpirun_cmd, (tuple, list)) or not all(isinstance(i, six.string_types) for i in mpirun_cmd):
            raise exceptions.ValidationError("the mpirun_command must be a list of strings")

        try:
            job_resource_keys = self.get_scheduler().job_resource_class.get_valid_keys()
        except exceptions.EntryPointError:
            raise exceptions.ValidationError("Unable to load the scheduler for this computer")

        subst = {i: 'value' for i in job_resource_keys}
        subst['tot_num_mpiprocs'] = 'value'

        try:
            for arg in mpirun_cmd:
                arg.format(**subst)
        except KeyError as exc:
            raise exceptions.ValidationError("In workdir there is an unknown replacement field {}".format(exc.args[0]))
        except ValueError as exc:
            raise exceptions.ValidationError("Error in the string: '{}'".format(exc))

    def validate(self):
        """
        Check if the attributes and files retrieved from the DB are valid.
        Raise a ValidationError if something is wrong.

        Must be able to work even before storing: therefore, use the get_attr and similar methods
        that automatically read either from the DB or from the internal attribute cache.

        For the base class, this is always valid. Subclasses will reimplement this.
        In the subclass, always call the super().validate() method first!
        """
        if not self.get_name().strip():
            raise exceptions.ValidationError("No name specified")

        self._hostname_validator(self.get_hostname())
        self._description_validator(self.get_description())
        self._transport_type_validator(self.get_transport_type())
        self._scheduler_type_validator(self.get_scheduler_type())
        self._workdir_validator(self.get_workdir())

        try:
            mpirun_cmd = self.get_mpirun_command()
        except exceptions.DbContentError:
            raise exceptions.ValidationError("Error in the DB content of the transport_params")

        # To be called AFTER the validation of the scheduler
        self._mpirun_command_validator(mpirun_cmd)

    @classmethod
    def _default_mpiprocs_per_machine_validator(cls, def_cpus_per_machine):
        """
        Validates the default number of CPUs per machine (node)
        """
        if def_cpus_per_machine is None:
            return

        if not isinstance(def_cpus_per_machine, six.integer_types) or def_cpus_per_machine <= 0:
            raise exceptions.ValidationError("Invalid value for default_mpiprocs_per_machine, "
                                             "must be a positive integer, or an empty "
                                             "string if you do not want to provide a "
                                             "default value.")

    # endregion

    def copy(self):
        """
        Return a copy of the current object to work with, not stored yet.
        """
        return Computer.from_backend_entity(self._backend_entity.copy())

    def store(self):
        """
        Store the computer in the DB.

        Differently from Nodes, a computer can be re-stored if its properties
        are to be changed (e.g. a new mpirun command, etc.)
        """
        self.validate()
        return super(Computer, self).store()

    @property
    def name(self):
        return self._backend_entity.name

    @property
    def label(self):
        """
        The computer label
        """
        return self.name

    @label.setter
    def label(self, value):
        """
        Set the computer label (i.e., name)
        """
        self.set_name(value)

    @property
    def description(self):
        """
        Get a description of the computer

        :return: the description
        :rtype: str
        """
        return self._backend_entity.description

    @property
    def hostname(self):
        return self._backend_entity.hostname

    def get_metadata(self):
        return self._backend_entity.get_metadata()

    def set_metadata(self, metadata):
        """
        Set the metadata.

        .. note: You still need to call the .store() method to actually save
           data to the database! (The store method can be called multiple
           times, differently from AiiDA Node objects).
        """
        self._backend_entity.set_metadata(metadata)

    def delete_property(self, name, raise_exception=True):
        """
        Delete a property from this computer

        :param name: the name of the property
        :param raise_exception: if True raise if the property does not exist, otherwise return None
        """
        olddata = self.get_metadata()
        try:
            del olddata[name]
            self.set_metadata(olddata)
        except KeyError:
            if raise_exception:
                raise AttributeError("'{}' property not found".format(name))

    def set_property(self, name, value):
        """
        Set a property on this computer

        :param name: the property name
        :param value: the new value
        """
        metadata = self.get_metadata() or {}
        metadata[name] = value
        self.set_metadata(metadata)

    def get_property(self, name, *args):
        """
        Get a property of this computer

        :param name: the property name
        :param args: additional arguments
        :return: the property value
        """
        if len(args) > 1:
            raise TypeError("get_property expected at most 2 arguments")
        olddata = self.get_metadata()
        try:
            return olddata[name]
        except KeyError:
            if not args:
                raise AttributeError("'{}' property not found".format(name))
            return args[0]

    def get_prepend_text(self):
        return self.get_property("prepend_text", "")

    def set_prepend_text(self, val):
        self.set_property("prepend_text", six.text_type(val))

    def get_append_text(self):
        return self.get_property("append_text", "")

    def set_append_text(self, val):
        self.set_property("append_text", six.text_type(val))

    def get_mpirun_command(self):
        """
        Return the mpirun command. Must be a list of strings, that will be
        then joined with spaces when submitting.

        I also provide a sensible default that may be ok in many cases.
        """
        return self.get_property("mpirun_command", ["mpirun", "-np", "{tot_num_mpiprocs}"])

    def set_mpirun_command(self, val):
        """
        Set the mpirun command. It must be a list of strings (you can use
        string.split() if you have a single, space-separated string).
        """
        if not isinstance(val, (tuple, list)) or not all(isinstance(i, six.string_types) for i in val):
            raise TypeError("the mpirun_command must be a list of strings")
        self.set_property("mpirun_command", val)

    def get_default_mpiprocs_per_machine(self):
        """
        Return the default number of CPUs per machine (node) for this computer,
        or None if it was not set.
        """
        return self.get_property("default_mpiprocs_per_machine", None)

    def set_default_mpiprocs_per_machine(self, def_cpus_per_machine):
        """
        Set the default number of CPUs per machine (node) for this computer.
        Accepts None if you do not want to set this value.
        """
        if def_cpus_per_machine is None:
            self.delete_property("default_mpiprocs_per_machine", raise_exception=False)
        else:
            if not isinstance(def_cpus_per_machine, six.integer_types):
                raise TypeError("def_cpus_per_machine must be an integer (or None)")
        self.set_property("default_mpiprocs_per_machine", def_cpus_per_machine)

    def get_minimum_job_poll_interval(self):
        """
        Get the minimum interval between subsequent requests to update the list
        of jobs currently running on this computer.

        :return: The minimum interval (in seconds)
        :rtype: float
        """
        return self.get_property(self.PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL,
                                 self.PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL__DEFAULT)

    def set_minimum_job_poll_interval(self, interval):
        """
        Set the minimum interval between subsequent requests to update the list
        of jobs currently running on this computer.

        :param interval: The minimum interval in seconds
        :type interval: float
        """
        self.set_property(self.PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL, interval)

    def get_transport_params(self):
        return self._backend_entity.get_transport_params()

    def set_transport_params(self, val):
        self._backend_entity.set_transport_params(val)

    def get_transport(self, user=None):
        """
        Return a Transport class, configured with all correct parameters.
        The Transport is closed (meaning that if you want to run any operation with
        it, you have to open it first (i.e., e.g. for a SSH transport, you have
        to open a connection). To do this you can call ``transports.open()``, or simply
        run within a ``with`` statement::

           transport = Computer.get_transport()
           with transport:
               print(transports.whoami())

        :param user: if None, try to obtain a transport for the default user.
            Otherwise, pass a valid User.

        :return: a (closed) Transport, already configured with the connection
            parameters to the supercomputer, as configured with ``verdi computer configure``
            for the user specified as a parameter ``user``.
        """
        from . import authinfos  # pylint: disable=cyclic-import

        user = user or users.User.objects(self.backend).get_default()
        authinfo = authinfos.AuthInfo.objects(self.backend).get(dbcomputer=self, aiidauser=user)
        return authinfo.get_transport()

    def get_workdir(self):
        """
        Get the working directory for this computer
        :return: The currently configured working directory
        :rtype: str
        """
        return self.get_property(self.PROPERTY_WORKDIR, "/scratch/{username}/aiida_run/")

    def set_workdir(self, val):
        self.set_property(self.PROPERTY_WORKDIR, val)

    def get_shebang(self):
        return self.get_property(self.PROPERTY_SHEBANG, "#!/bin/bash")

    def set_shebang(self, val):
        """
        :param str val: A valid shebang line
        """
        if not isinstance(val, six.string_types):
            raise ValueError("{} is invalid. Input has to be a string".format(val))
        if not val.startswith('#!'):
            raise ValueError("{} is invalid. A shebang line has to start with #!".format(val))
        metadata = self.get_metadata()
        metadata['shebang'] = val
        self.set_metadata(metadata)

    def get_name(self):
        return self._backend_entity.get_name()

    def set_name(self, val):
        self._backend_entity.set_name(val)

    def get_hostname(self):
        """
        Get this computer hostname
        :rtype: str
        """
        return self._backend_entity.get_hostname()

    def set_hostname(self, val):
        """
        Set the hostname of this computer
        :param val: The new hostname
        :type val: str
        """
        self._backend_entity.set_hostname(val)

    def get_description(self):
        """
        Get the description for this computer

        :return: the description
        :rtype: str
        """

    def set_description(self, val):
        """
        Set the description for this computer

        :param val: the new description
        :type val: str
        """
        self._backend_entity.set_description(val)

    def get_authinfo(self, user):
        """
        Return the aiida.orm.authinfo.AuthInfo instance for the
        given user on this computer, if the computer
        is configured for the given user.

        :param user: a User instance.
        :return: a AuthInfo instance
        :raise aiida.common.NotExistent: if the computer is not configured for the given
            user.
        """
        from . import authinfos

        return authinfos.AuthInfo.objects(self.backend).get(dbcomputer_id=self.id, aiidauser_id=user.id)

    def is_user_configured(self, user):
        """
        Is the user configured on this computer?

        :param user: the user to check
        :return: True if configured, False otherwise
        :rtype: bool
        """
        try:
            self.get_authinfo(user)
            return True
        except exceptions.NotExistent:
            return False

    def is_user_enabled(self, user):
        """
        Is the given user enabled to run on this computer?

        :param user: the user to check
        :return: True if enabled, False otherwise
        :rtype: bool
        """
        try:
            authinfo = self.get_authinfo(user)
            return authinfo.enabled
        except exceptions.NotExistent:
            # Return False if the user is not configured (in a sense,
            # it is disabled for that user)
            return False

    def get_scheduler_type(self):
        """
        Get the scheduler type for this computer

        :return: the scheduler type
        :rtype: str
        """
        return self._backend_entity.get_scheduler_type()

    def set_scheduler_type(self, scheduler_type):
        """
        :param scheduler_type: the new scheduler type
        """
        self._scheduler_type_validator(scheduler_type)
        self._backend_entity.set_scheduler_type(scheduler_type)

    def get_transport_type(self):
        """
        Get the current transport type for this computer

        :return: the transport type
        :rtype: str
        """
        return self._backend_entity.get_transport_type()

    def set_transport_type(self, transport_type):
        """
        Set the transport type for this computer

        :param transport_type: the new transport type
        :type transport_type: str
        """
        self._backend_entity.set_transport_type(transport_type)

    def get_transport_class(self):
        """
        Get the transport class for this computer.  Can be used to instantiate a transport instance.

        :return: the transport class
        """
        try:
            return TransportFactory(self.get_transport_type())
        except exceptions.EntryPointError as exception:
            raise exceptions.ConfigurationError('No transport found for {} [type {}], message: {}'.format(
                self.name, self.get_transport_type(), exception))

    def get_scheduler(self):
        """
        Get a scheduler instance for this computer

        :return: the scheduler instance
        :rtype: :class:`aiida.schedulers.Scheduler`
        """
        try:
            scheduler_class = SchedulerFactory(self.get_scheduler_type())
            # I call the init without any parameter
            return scheduler_class()
        except exceptions.EntryPointError as exception:
            raise exceptions.ConfigurationError('No scheduler found for {} [type {}], message: {}'.format(
                self.name, self.get_scheduler_type(), exception))

    def configure(self, user=None, **kwargs):
        """
        Configure a computer for a user with valid auth params passed via kwargs

        :param user: the user to configure the computer for
        :kwargs: the configuration keywords with corresponding values
        :return: the authinfo object for the configured user
        :rtype: :class:`aiida.orm.AuthInfo`
        """
        from . import authinfos

        transport_cls = self.get_transport_class()
        user = user or users.User.objects(self.backend).get_default()
        valid_keys = set(transport_cls.get_valid_auth_params())

        if not set(kwargs.keys()).issubset(valid_keys):
            invalid_keys = [key for key in kwargs if key not in valid_keys]
            raise ValueError('{transport}: received invalid authentication parameter(s) "{invalid}"'.format(
                transport=transport_cls, invalid=invalid_keys))

        try:
            authinfo = self.get_authinfo(user)
        except exceptions.NotExistent:
            authinfo = authinfos.AuthInfo(self, user)

        auth_params = authinfo.get_auth_params()

        if valid_keys:
            auth_params.update(kwargs)
            authinfo.set_auth_params(auth_params)
            authinfo.store()

        return authinfo

    def get_configuration(self, user=None):
        """
        Get the configuration of computer for the given user as a dictionary

        :param user: the user to to get the configuration for.  Uses default user if `None`
        :type user: :class:`aiida.orm.User`
        """

        backend = self.backend
        user = user or users.User.objects(self.backend).get_default()

        config = {}
        try:
            authinfo = backend.authinfos.get(self, user)
            config = authinfo.get_auth_params()
        except exceptions.NotExistent:
            pass

        return config

    @staticmethod
    def get_schema():
        """
        Every node property contains:
            - display_name: display name of the property
            - help text: short help text of the property
            - is_foreign_key: is the property foreign key to other type of the node
            - type: type of the property. e.g. str, dict, int

        :return: get schema of the computer
        """
        return {
            "description": {
                "display_name": "Description",
                "help_text": "short description of the Computer",
                "is_foreign_key": False,
                "type": "str"
            },
            "hostname": {
                "display_name": "Host",
                "help_text": "Name of the host",
                "is_foreign_key": False,
                "type": "str"
            },
            "id": {
                "display_name": "Id",
                "help_text": "Id of the object",
                "is_foreign_key": False,
                "type": "int"
            },
            "name": {
                "display_name": "Name",
                "help_text": "Name of the object",
                "is_foreign_key": False,
                "type": "str"
            },
            "scheduler_type": {
                "display_name": "Scheduler",
                "help_text": "Scheduler type",
                "is_foreign_key": False,
                "type": "str",
                "valid_choices": {
                    "direct": {
                        "doc": "Support for the direct execution bypassing schedulers."
                    },
                    "pbsbaseclasses.PbsBaseClass": {
                        "doc": "Base class with support for the PBSPro scheduler"
                    },
                    "pbspro": {
                        "doc": "Subclass to support the PBSPro scheduler"
                    },
                    "sge": {
                        "doc":
                        "Support for the Sun Grid Engine scheduler and its variants/forks (Son of Grid Engine, "
                        "Oracle Grid Engine, ...)"
                    },
                    "slurm": {
                        "doc": "Support for the SLURM scheduler (http://slurm.schedmd.com/)."
                    },
                    "torque": {
                        "doc": "Subclass to support the Torque scheduler."
                    }
                }
            },
            "transport_params": {
                "display_name": "",
                "help_text": "Transport Parameters",
                "is_foreign_key": False,
                "type": "str"
            },
            "transport_type": {
                "display_name": "Transport type",
                "help_text": "Transport Type",
                "is_foreign_key": False,
                "type": "str",
                "valid_choices": {
                    "local": {
                        "doc":
                        "Support copy and command execution on the same host on which AiiDA is running via direct file "
                        "copy and execution commands."
                    },
                    "ssh": {
                        "doc":
                        "Support connection, command execution and data transfer to remote computers via SSH+SFTP."
                    }
                }
            },
            "uuid": {
                "display_name": "Unique ID",
                "help_text": "Universally Unique Identifier",
                "is_foreign_key": False,
                "type": "unicode"
            }
        }



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import io
import os

from aiida.common import exceptions
from aiida.orm import Dict
from aiida.parsers.parser import Parser
from aiida.plugins import CalculationFactory

TemplatereplacerCalculation = CalculationFactory('templatereplacer')


class TemplatereplacerDoublerParser(Parser):

    def parse(self, **kwargs):
        """Parse the contents of the output files retrieved in the `FolderData`."""
        template = self.node.inputs.template.get_dict()

        try:
            output_folder = self.retrieved
        except exceptions.NotExistent:
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        try:
            output_file = template['output_file_name']
        except KeyError:
            return self.exit_codes.ERROR_NO_OUTPUT_FILE_NAME_DEFINED

        try:
            with output_folder.open(output_file, 'r') as handle:
                result = self.parse_stdout(handle)
        except (OSError, IOError):
            self.logger.exception('unable to parse the output for CalcJobNode<{}>'.format(self.node.pk))
            return self.exit_codes.ERROR_READING_OUTPUT_FILE

        output_dict = {'value': result, 'retrieved_temporary_files': []}
        retrieve_temporary_files = template.get('retrieve_temporary_files', None)

        # If the 'retrieve_temporary_files' key was set in the template input node, we expect a temporary directory
        # to have been passed in the keyword arguments under the name `retrieved_temporary_folder`.
        if retrieve_temporary_files is not None:
            try:
                retrieved_temporary_folder = kwargs['retrieved_temporary_folder']
            except KeyError:
                return self.exit_codes.ERROR_NO_TEMPORARY_RETRIEVED_FOLDER

            for retrieved_file in retrieve_temporary_files:

                file_path = os.path.join(retrieved_temporary_folder, retrieved_file)

                if not os.path.isfile(file_path):
                    self.logger.error('the file {} was not found in the temporary retrieved folder {}'.format(
                        retrieved_file, retrieved_temporary_folder))
                    return self.exit_codes.ERROR_READING_TEMPORARY_RETRIEVED_FILE

                with io.open(file_path, 'r', encoding='utf8') as handle:
                    parsed_value = handle.read().strip()

                # We always strip the content of the file from whitespace to simplify testing for expected output
                output_dict['retrieved_temporary_files'].append((retrieved_file, parsed_value))

        self.out(self.node.process_class.spec().default_output_node, Dict(dict=output_dict))

        return

    @staticmethod
    def parse_stdout(filelike):
        """
        Parse the sum from the output of the ArithmeticAddcalculation written to standard out

        :param filelike: filelike object containing the output
        :returns: the sum
        """
        try:
            result = int(filelike.read())
        except ValueError:
            result = None

        return result



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for Path types"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.params.types import ImportPath


class TestImportPath(AiidaTestCase):
    """Tests `ImportPath`"""

    def test_default_timeout(self):
        """Test the default timeout_seconds value is correct"""
        from aiida.cmdline.params.types.path import URL_TIMEOUT_SECONDS

        import_path = ImportPath()

        self.assertEqual(import_path.timeout_seconds, URL_TIMEOUT_SECONDS)

    def test_valid_timeout(self):
        """Test a valid timeout_seconds value"""

        valid_values = [42, "42"]

        for value in valid_values:
            import_path = ImportPath(timeout_seconds=value)

            self.assertEqual(import_path.timeout_seconds, int(value))

    def test_none_timeout(self):
        """Test a TypeError is raised when a None value is given for timeout_seconds"""

        with self.assertRaises(TypeError):
            ImportPath(timeout_seconds=None)

    def test_wrong_type_timeout(self):
        """Test a TypeError is raised when wrong type is given for timeout_seconds"""

        with self.assertRaises(TypeError):
            ImportPath(timeout_seconds="test")

    def test_range_timeout(self):
        """Test timeout_seconds defines extrema when out of range
        Range of timeout_seconds is [0;60], extrema included.
        """

        range_timeout = [0, 60]
        lower = range_timeout[0] - 5
        upper = range_timeout[1] + 5

        lower_path = ImportPath(timeout_seconds=lower)
        upper_path = ImportPath(timeout_seconds=upper)

        msg_lower = "timeout_seconds should have been corrected to the lower bound: '{}', but instead it is {}".format(
            range_timeout[0], lower_path.timeout_seconds)
        self.assertEqual(lower_path.timeout_seconds, range_timeout[0], msg_lower)

        msg_upper = "timeout_seconds should have been corrected to the upper bound: '{}', but instead it is {}".format(
            range_timeout[1], upper_path.timeout_seconds)
        self.assertEqual(upper_path.timeout_seconds, range_timeout[1], msg_upper)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility functions for command line commands related to the daemon."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click
from tabulate import tabulate

from aiida.cmdline.utils.common import format_local_time


def print_client_response_status(response):
    """
    Print the response status of a call to the CircusClient through the DaemonClient

    :param response: the response object
    """
    from aiida.engine.daemon.client import DaemonClient

    if 'status' not in response:
        return

    if response['status'] == 'active':
        click.secho('RUNNING', fg='green', bold=True)
    elif response['status'] == 'ok':
        click.secho('OK', fg='green', bold=True)
    elif response['status'] == DaemonClient.DAEMON_ERROR_NOT_RUNNING:
        click.secho('FAILED', fg='red', bold=True)
        click.echo('Try to run \'verdi daemon start --foreground\' to potentially see the exception')
    elif response['status'] == DaemonClient.DAEMON_ERROR_TIMEOUT:
        click.secho('TIMEOUT', fg='red', bold=True)
    else:
        click.echo(response['status'])


def get_daemon_status(client):
    """
    Print the status information of the daemon for a given profile through its DaemonClient

    :param client: the DaemonClient
    """

    if not client.is_daemon_running:
        return 'The daemon is not running'

    status_response = client.get_status()

    if status_response['status'] == 'stopped':
        return 'The daemon is paused'

    if status_response['status'] == 'error':
        return 'The daemon is in an unexpected state, try verdi daemon restart --reset'

    if status_response['status'] == 'timeout':
        return 'The daemon is running but the call to the circus controller timed out'

    worker_response = client.get_worker_info()
    daemon_response = client.get_daemon_info()

    if 'info' not in worker_response or 'info' not in daemon_response:
        return 'Call to the circus controller timed out'

    workers = [['PID', 'MEM %', 'CPU %', 'started']]
    for worker_pid, worker_info in worker_response['info'].items():
        worker_row = [worker_pid, worker_info['mem'], worker_info['cpu'], format_local_time(worker_info['create_time'])]
        workers.append(worker_row)

    if len(workers) > 1:
        workers_info = tabulate(workers, headers='firstrow', tablefmt='simple')
    else:
        workers_info = '--> No workers are running. Use verdi daemon incr to start some!\n'

    info = {
        'pid': daemon_response['info']['pid'],
        'time': format_local_time(daemon_response['info']['create_time']),
        'nworkers': len(workers) - 1,
        'workers': workers_info
    }

    template = ('Daemon is running as PID {pid} since {time}\nActive workers [{nworkers}]:\n{workers}\n'
                'Use verdi daemon [incr | decr] [num] to increase / decrease the amount of workers')

    return template.format(**info)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Definition of known configuration options and methods to parse and get option values."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import io
import os

from .config import Config
from .migrations import check_and_migrate_config
from .settings import DEFAULT_CONFIG_FILE_NAME

__all__ = ('load_config',)


def load_config(create=False):
    """Instantiate the Config object representing the configuration file of the current AiiDA instance.

    :param create: when set to True, will create the configuration file if it does not already exist
    :return: the config
    :rtype: :class:`~aiida.manage.configuration.config.Config`
    :raises aiida.common.MissingConfigurationError: if the configuration file could not be found and create=False
    """
    from .settings import AIIDA_CONFIG_FOLDER
    from aiida.backends.settings import IN_RT_DOC_MODE, DUMMY_CONF_FILE
    from aiida.common import exceptions
    from aiida.common import json

    if IN_RT_DOC_MODE:
        return DUMMY_CONF_FILE

    filepath = os.path.join(AIIDA_CONFIG_FOLDER, DEFAULT_CONFIG_FILE_NAME)

    if not os.path.isfile(filepath):
        if not create:
            raise exceptions.MissingConfigurationError('configuration file {} does not exist'.format(filepath))
        else:
            config = Config(filepath, {}).store()
    else:
        try:
            with io.open(filepath, 'r', encoding='utf8') as handle:
                config = Config(filepath, json.load(handle))
        except ValueError:
            raise exceptions.ConfigurationError('configuration file {} contains invalid JSON'.format(filepath))

    return check_and_migrate_config(config, store=True)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data structure` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from six.moves import range

import click

from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.commands.cmd_data import cmd_show
from aiida.cmdline.commands.cmd_data.cmd_export import data_export, export_options
from aiida.cmdline.commands.cmd_data.cmd_list import data_list, list_options
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo

LIST_PROJECT_HEADERS = ['Id', 'Label', 'Kinds', 'Sites']
EXPORT_FORMATS = ['cif', 'xsf', 'xyz']
IMPORT_FORMATS = ['ase', 'pwi', 'xyz']
VISUALIZATION_FORMATS = ['ase', 'jmol', 'vesta', 'vmd', 'xcrysden']


@verdi_data.group('structure')
def structure():
    """Manipulation of StructureData objects."""


# pylint: disable=too-many-locals,too-many-branches
@structure.command('list')
@options.FORMULA_MODE()
@options.WITH_ELEMENTS()
@list_options
@decorators.with_dbenv()
def structure_list(elements, raw, formula_mode, past_days, groups, all_users):
    """List stored StructureData objects."""
    from aiida.orm.nodes.data.structure import StructureData, get_formula, get_symbols_string
    from tabulate import tabulate

    elements_only = False
    lst = data_list(StructureData, LIST_PROJECT_HEADERS, elements, elements_only, formula_mode, past_days, groups,
                    all_users)

    entry_list = []
    for [pid, label, akinds, asites] in lst:
        # If symbols are defined there is a filtering of the structures
        # based on the element
        # When QueryBuilder will support this (attribute)s filtering,
        # it will be pushed in the query.
        if elements is not None:
            all_symbols = [_["symbols"][0] for _ in akinds]
            if not any([s in elements for s in all_symbols]):
                continue

            if elements_only:
                echo.echo_critical("Not implemented elements-only search")

        # We want only the StructureData that have attributes
        if akinds is None or asites is None:
            continue

        symbol_dict = {}
        for k in akinds:
            symbols = k['symbols']
            weights = k['weights']
            symbol_dict[k['name']] = get_symbols_string(symbols, weights)

        try:
            symbol_list = []
            for site in asites:
                symbol_list.append(symbol_dict[site['kind_name']])
            formula = get_formula(symbol_list, mode=formula_mode)
        # If for some reason there is no kind with the name
        # referenced by the site
        except KeyError:
            formula = "<<UNKNOWN>>"
        entry_list.append([str(pid), str(formula), label])

    counter = 0
    struct_list_data = list()
    if not raw:
        struct_list_data.append(LIST_PROJECT_HEADERS)
    for entry in entry_list:
        for i, value in enumerate(entry):
            if isinstance(value, list):
                entry[i] = ",".join(value)
        for i in range(len(entry), len(LIST_PROJECT_HEADERS)):
            entry.append(None)
        counter += 1
    struct_list_data.extend(entry_list)
    if raw:
        echo.echo(tabulate(struct_list_data, tablefmt='plain'))
    else:
        echo.echo(tabulate(struct_list_data, headers="firstrow"))
        echo.echo("\nTotal results: {}\n".format(counter))


@structure.command('show')
@arguments.DATA(type=types.DataParamType(sub_classes=('aiida.data:structure',)))
@options.VISUALIZATION_FORMAT(type=click.Choice(VISUALIZATION_FORMATS), default='ase')
@decorators.with_dbenv()
def structure_show(data, fmt):
    """Visualize StructureData objects."""
    try:
        show_function = getattr(cmd_show, '_show_{}'.format(fmt))
    except AttributeError:
        echo.echo_critical('visualization format {} is not supported'.format(fmt))

    show_function(fmt, data)


@structure.command('export')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:structure',)))
@options.EXPORT_FORMAT(
    type=click.Choice(EXPORT_FORMATS),
    default='xyz',
)
@export_options
@decorators.with_dbenv()
def structure_export(**kwargs):
    """Export StructureData object."""
    node = kwargs.pop('datum')
    output = kwargs.pop('output')
    fmt = kwargs.pop('fmt')
    force = kwargs.pop('force')

    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    data_export(node, output, fmt, other_args=kwargs, overwrite=force)


# pylint: disable=too-many-arguments
@structure.command('import')
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@options.INPUT_FORMAT(type=click.Choice(IMPORT_FORMATS))
@click.option(
    '--vacuum-factor',
    type=click.FLOAT,
    show_default=True,
    default=1.0,
    help="The factor by which the cell accomodating the structure should be increased.")
@click.option(
    '--vacuum-addition',
    type=click.FLOAT,
    show_default=True,
    default=10.0,
    help='The distance to add to the unit cell after vacuum factor was applied to expand in each dimension')
@click.option(
    '--pbc',
    type=click.INT,
    nargs=3,
    show_default=True,
    default=[0, 0, 0],
    help='Set periodic boundary conditions for each lattice direction, where 0 means no periodicity')
@click.option('--view', is_flag=True, default=False, help='View resulting structure using ASE.')
@click.option('--store', is_flag=True, default=False, help='Do not store the structure in AiiDA database.')
@decorators.with_dbenv()
def structure_import(filename, fmt, vacuum_factor, vacuum_addition, pbc, view, store):
    """
    Import structure
    """
    from aiida.cmdline.commands.cmd_data import cmd_import

    args = {
        'vacuum_factor': vacuum_factor,
        'vacuum_addition': vacuum_addition,
        'pbc': pbc,
        'view': view,
        'store': store,
    }

    try:
        show_function = getattr(cmd_import, 'data_import_{}'.format(fmt))
    except AttributeError:
        echo.echo_critical('import format {} is not supported'.format(fmt))

    show_function(filename, **args)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Testing tools for related projects like plugins.

Fixtures (pytest) and specific test classes & test runners (unittest)
that set up a complete temporary AiiDA environment for plugin tests.

Filesystem:

    * temporary config (``.aiida``) folder
    * temporary repository folder

Database:

    * temporary database cluster via the ``pgtest`` package
    * with aiida database user
    * with aiida_db database

AiiDA:

    * set to use the temporary config folder
    * create and configure a profile

"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
import unittest
import tempfile
import shutil
from os import path
from contextlib import contextmanager

from pgtest.pgtest import PGTest

from aiida import is_dbenv_loaded
from aiida.backends import settings as backend_settings
from aiida.backends.profile import BACKEND_DJANGO, BACKEND_SQLA
from aiida.common import exceptions
from aiida.manage.manager import get_manager, reset_manager
from aiida.manage.configuration.setup import create_instance_directories
from aiida.manage.configuration.utils import load_config
from aiida.manage.external.postgres import Postgres


class FixtureError(Exception):
    """Raised by FixtureManager, when it encounters a situation in which consistent behaviour can not be guaranteed"""

    def __init__(self, msg):
        super(FixtureError, self).__init__()
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class FixtureManager(object):  # pylint: disable=too-many-public-methods,useless-object-inheritance
    """
    Manage the life cycle of a completely separated and temporary AiiDA environment

    * No previously created database of profile is required to run tests using this environment
    * Tests using this environment will never pollute the user's work environment

    Example::

        fixtures = FixtureManager()
        fixtures.create_aiida_db()  # set up only the database
        fixtures.create_profile()  # set up a profile (creates the db too if necessary)

        # ready for testing

        # run test 1

        fixtures.reset_db()
        # database ready for independent test 2

        # run test 2

        fixtures.destroy_all()
        # everything cleaned up

    Usage (unittest): See the :py:class:`PluginTestCase` and the :py:class:`TestRunner`.


    Usage (pytest)::

        import pytest

        @pytest.fixture(scope='session')
        def aiida_profile():
            # set up a test profile for the duration of the tests
            with aiida.manage.fixtures.fixture_manager() as fixture_manager:
                yield fixture_manager

        @pytest.fixture(scope='function')
        def new_database(aiida_profile):
            # clear the database after each test
            yield aiida_profile
            aiida_profile.reset_db()

        def test_my_stuff(new_database):
            # run a test

    """

    _test_case = None

    def __init__(self):
        from aiida.manage import configuration
        from aiida.manage.configuration import settings as configuration_settings
        self.db_params = {}
        self.fs_env = {'repo': 'test_repo', 'config': '.aiida'}
        self.profile_info = {
            'backend': 'django',
            'email': 'test@aiida.mail',
            'first_name': 'AiiDA',
            'last_name': 'Plugintest',
            'institution': 'aiidateam',
            'db_user': 'aiida',
            'db_pass': 'aiida_pw',
            'db_name': 'aiida_db'
        }
        self.pg_cluster = None
        self.postgres = None
        self.__is_running_on_test_db = False
        self.__is_running_on_test_profile = False
        self._backup = {}
        self._backup['config'] = configuration.CONFIG
        self._backup['config_dir'] = configuration_settings.AIIDA_CONFIG_FOLDER
        self._backup['profile'] = backend_settings.AIIDADB_PROFILE
        self.__backend = None

    @property
    def _backend(self):
        """
        Get the backend
        """
        if self.__backend is None:
            # Lazy load the backend so we don't do it too early (i.e. before load_dbenv())
            self.__backend = get_manager().get_backend()
        return self.__backend

    def create_db_cluster(self):
        if not self.pg_cluster:
            self.pg_cluster = PGTest()
        self.db_params.update(self.pg_cluster.dsn)

    def create_aiida_db(self):
        """Create the necessary database on the temporary postgres instance"""
        if is_dbenv_loaded():
            raise FixtureError('AiiDA dbenv can not be loaded while creating a test db environment')
        if not self.db_params:
            self.create_db_cluster()
        self.postgres = Postgres(interactive=False, quiet=True, dbinfo=self.db_params)
        self.postgres.determine_setup()
        self.db_params = self.postgres.get_dbinfo()
        if not self.postgres.pg_execute:
            raise FixtureError('Could not connect to the test postgres instance')
        self.postgres.create_dbuser(self.db_user, self.db_pass)
        self.postgres.create_db(self.db_user, self.db_name)
        self.__is_running_on_test_db = True

    def create_profile(self):
        """
        Set AiiDA to use the test config dir and create a default profile there

        Warning: the AiiDA dbenv must not be loaded when this is called!
        """
        if is_dbenv_loaded():
            raise FixtureError('AiiDA dbenv can not be loaded while creating a test profile')
        if not self.__is_running_on_test_db:
            self.create_aiida_db()
        from aiida.manage import configuration
        from aiida.manage.configuration import settings as configuration_settings
        from aiida.manage.configuration.setup import setup_profile
        if not self.root_dir:
            self.root_dir = tempfile.mkdtemp()
        configuration.CONFIG = None
        configuration_settings.AIIDA_CONFIG_FOLDER = self.config_dir
        backend_settings.AIIDADB_PROFILE = None
        create_instance_directories()
        config = load_config(create=True)
        profile_name = 'test_profile'
        setup_profile(profile_name=profile_name, only_config=False, non_interactive=True, **self.profile)
        config = load_config()
        config.set_default_profile(profile_name).store()
        self.__is_running_on_test_profile = True
        self._create_test_case()
        self.init_db()

    def reset_db(self):
        """Cleans all data from the database between tests"""

        self._test_case.clean_db()
        reset_manager()
        self.init_db()

    @staticmethod
    def init_db():
        """Initialise the database state"""
        # Create the default user
        from aiida import orm
        try:
            orm.User(email=get_manager().get_profile().default_user_email).store()
        except exceptions.IntegrityError:
            # The default user already exists, no problem
            pass

    @property
    def profile(self):
        """Profile parameters"""
        profile = {
            'backend': self.backend,
            'email': self.email,
            'repo': self.repo,
            'db_host': self.db_host,
            'db_port': self.db_port,
            'db_name': self.db_name,
            'db_user': self.db_user,
            'db_pass': self.db_pass,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'institution': self.institution
        }
        return profile

    @property
    def db_host(self):
        return self.db_params.get('host')

    @db_host.setter
    def db_host(self, hostname):
        self.db_params['host'] = hostname

    @property
    def first_name(self):
        return self.profile_info['first_name']

    @first_name.setter
    def first_name(self, name):
        self.profile_info['first_name'] = name

    @property
    def last_name(self):
        return self.profile_info['last_name']

    @last_name.setter
    def last_name(self, name):
        self.profile_info['last_name'] = name

    @property
    def institution(self):
        return self.profile_info['institution']

    @institution.setter
    def institution(self, institution):
        self.profile_info['institution'] = institution

    @property
    def db_port(self):
        return self.db_params.get('port', None)

    @db_port.setter
    def db_port(self, port):
        self.db_params['port'] = str(port)

    def repo_ok(self):
        return bool(self.repo and path.isdir(path.dirname(self.repo)))

    @property
    def repo(self):
        return self._return_dir('repo')

    @repo.setter
    def repo(self, repo_dir):
        self.fs_env['repo'] = repo_dir

    def _return_dir(self, key):
        """Return a path to a directory from the fs environment"""
        dir_path = self.fs_env[key]
        if not dir_path:
            raise FixtureError('no directory set for {}'.format(key))
        elif path.isabs(dir_path):
            return dir_path
        return path.join(self.root_dir, dir_path)

    @property
    def email(self):
        return self.profile_info['email']

    @email.setter
    def email(self, email):
        self.profile_info['email'] = email

    @property
    def backend(self):
        return self.profile_info['backend']

    @backend.setter
    def backend(self, backend):
        if self.__is_running_on_test_profile:
            raise FixtureError('backend cannot be changed after setting up the environment')

        valid_backends = [BACKEND_DJANGO, BACKEND_SQLA]
        if backend not in valid_backends:
            raise ValueError('invalid backend {}, must be one of {}'.format(backend, valid_backends))
        self.profile_info['backend'] = backend

    @property
    def config_dir_ok(self):
        return bool(self.config_dir and path.isdir(self.config_dir))

    @property
    def config_dir(self):
        return self._return_dir('config')

    @config_dir.setter
    def config_dir(self, config_dir):
        self.fs_env['config'] = config_dir

    @property
    def db_user(self):
        return self.profile_info['db_user']

    @db_user.setter
    def db_user(self, user):
        self.profile_info['db_user'] = user

    @property
    def db_pass(self):
        return self.profile_info['db_pass']

    @db_pass.setter
    def db_pass(self, passwd):
        self.profile_info['db_pass'] = passwd

    @property
    def db_name(self):
        return self.profile_info['db_name']

    @db_name.setter
    def db_name(self, name):
        self.profile_info['db_name'] = name

    @property
    def root_dir(self):
        return self.fs_env.get('root', '')

    @root_dir.setter
    def root_dir(self, root_dir):
        self.fs_env['root'] = root_dir

    @property
    def root_dir_ok(self):
        return bool(self.root_dir and path.isdir(self.root_dir))

    def destroy_all(self):
        """Remove all traces of the test run"""
        from aiida.manage import configuration
        from aiida.manage.configuration import settings as configuration_settings
        if self.root_dir:
            shutil.rmtree(self.root_dir)
            self.root_dir = None
        if self.pg_cluster:
            self.pg_cluster.close()
            self.pg_cluster = None
        self.__is_running_on_test_db = False
        self.__is_running_on_test_profile = False
        if 'config' in self._backup:
            configuration.CONFIG = self._backup['config']
        if 'config_dir' in self._backup:
            configuration_settings.AIIDA_CONFIG_FOLDER = self._backup['config_dir']
        if 'profile' in self._backup:
            backend_settings.AIIDADB_PROFILE = self._backup['profile']

    def _create_test_case(self):
        """
        Create the test case for the correct backend which will be used to clean up
        """
        if not self.__is_running_on_test_profile:
            raise FixtureError('No test profile has been set up yet, cannot create appropriate test case')
        if self.profile_info['backend'] == BACKEND_DJANGO:
            from aiida.backends.djsite.db.testbase import DjangoTests
            self._test_case = DjangoTests()
        elif self.profile_info['backend'] == BACKEND_SQLA:
            from aiida.backends.sqlalchemy.tests.testbase import SqlAlchemyTests
            from aiida.backends.sqlalchemy import get_scoped_session

            self._test_case = SqlAlchemyTests()
            self._test_case.test_session = get_scoped_session()

    def has_profile_open(self):
        return self.__is_running_on_test_profile


_GLOBAL_FIXTURE_MANAGER = FixtureManager()


@contextmanager
def fixture_manager(backend=BACKEND_DJANGO):
    """
    Context manager for FixtureManager objects

    Example test runner (unittest)::

        with fixture_manager(backend) as fixture_mgr:
            # ready for tests
        # everything cleaned up

    Example fixture (pytest)::

        def aiida_profile():
            with fixture_manager(backend) as fixture_mgr:
                yield fixture_mgr

    :param backend: database backend, either BACKEND_SQLA or BACKEND_DJANGO
    """
    try:
        if not _GLOBAL_FIXTURE_MANAGER.has_profile_open():
            _GLOBAL_FIXTURE_MANAGER.backend = backend
            _GLOBAL_FIXTURE_MANAGER.create_profile()
        yield _GLOBAL_FIXTURE_MANAGER
    finally:
        _GLOBAL_FIXTURE_MANAGER.destroy_all()


class PluginTestCase(unittest.TestCase):
    """
    Set up a complete temporary AiiDA environment for plugin tests.

    Note: This test class needs to be run through the :py:class:`TestRunner`
    and will **not** work simply with `python -m unittest discover`.

    Usage example::

        MyTestCase(aiida.manage.fixtures.PluginTestCase):

            def setUp(self):
                # load my test data

            # optionally extend setUpClass / tearDownClass / tearDown if needed

            def test_my_plugin(self):
                # execute tests
    """
    # Filled in during setUpClass
    backend = None  # type :class:`aiida.orm.Backend`

    @classmethod
    def setUpClass(cls):
        cls.fixture_manager = _GLOBAL_FIXTURE_MANAGER
        if not cls.fixture_manager.has_profile_open():
            raise ValueError(
                "Fixture mananger has no open profile. Please use aiida.manage.fixtures.TestRunner to run these tests.")

        cls.backend = get_manager().get_backend()

    def tearDown(self):
        self.fixture_manager.reset_db()


class TestRunner(unittest.runner.TextTestRunner):
    """
    Testrunner for unit tests using the fixture manager.

    Usage example::

        import unittest
        from aiida.manage.fixtures import TestRunner

        tests = unittest.defaultTestLoader.discover('.')
        TestRunner().run(tests)

    """

    # pylint: disable=arguments-differ
    def run(self, suite, backend=BACKEND_DJANGO):
        """
        Run tests using fixture manager for specified backend.

        :param tests: A suite of tests, as returned e.g. by :py:meth:`unittest.TestLoader.discover`
        :param backend: Database backend to be used.
        """
        from aiida.common.utils import Capturing
        with Capturing():
            with fixture_manager(backend=backend):
                return super(TestRunner, self).run(suite)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Serialization tests"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida import orm
from aiida.orm.utils import serialize
from aiida.backends.testbase import AiidaTestCase

# pylint: disable=missing-docstring


class TestSerialize(AiidaTestCase):

    def test_serialize_round_trip(self):
        """
        Test the serialization of a dictionary with Nodes in various data structure
        Also make sure that the serialized data is json-serializable
        """
        node_a = orm.Data().store()
        node_b = orm.Data().store()

        data = {'test': 1, 'list': [1, 2, 3, node_a], 'dict': {('Si',): node_b, 'foo': 'bar'}, 'baz': 'aar'}

        serialized_data = serialize.serialize(data)
        deserialized_data = serialize.deserialize(serialized_data)

        # For now manual element-for-element comparison until we come up with general
        # purpose function that can equate two node instances properly
        self.assertEqual(data['test'], deserialized_data['test'])
        self.assertEqual(data['baz'], deserialized_data['baz'])
        self.assertEqual(data['list'][:3], deserialized_data['list'][:3])
        self.assertEqual(data['list'][3].uuid, deserialized_data['list'][3].uuid)
        self.assertEqual(data['dict'][('Si',)].uuid, deserialized_data['dict'][('Si',)].uuid)

    def test_serialize_group(self):
        """
        Test that serialization and deserialization of Groups works.
        Also make sure that the serialized data is json-serializable
        """
        group_name = 'groupie'
        group_a = orm.Group(label=group_name).store()

        data = {'group': group_a}

        serialized_data = serialize.serialize(data)
        deserialized_data = serialize.deserialize(serialized_data)

        self.assertEqual(data['group'].uuid, deserialized_data['group'].uuid)
        self.assertEqual(data['group'].label, deserialized_data['group'].label)

    def test_serialize_node_round_trip(self):
        """Test you can serialize and deserialize a node"""
        node = orm.Data().store()
        deserialized = serialize.deserialize(serialize.serialize(node))
        self.assertEqual(node.uuid, deserialized.uuid)

    def test_serialize_group_round_trip(self):
        """Test you can serialize and deserialize a group"""
        group = orm.Group(label='test_serialize_group_round_trip').store()
        deserialized = serialize.deserialize(serialize.serialize(group))

        self.assertEqual(group.uuid, deserialized.uuid)
        self.assertEqual(group.label, deserialized.label)

    def test_serialize_computer_round_trip(self):
        """Test you can serialize and deserialize a computer"""
        computer = self.computer
        deserialized = serialize.deserialize(serialize.serialize(computer))

        # pylint: disable=no-member
        self.assertEqual(computer.uuid, deserialized.uuid)
        self.assertEqual(computer.name, deserialized.name)

    def test_serialize_unstored_node(self):
        """Test that you can't serialize an unstored node"""
        node = orm.Data()

        with self.assertRaises(ValueError):
            serialize.serialize(node)

    def test_serialize_unstored_group(self):
        """Test that you can't serialize an unstored group"""
        group = orm.Group(label='test_serialize_unstored_group')

        with self.assertRaises(ValueError):
            serialize.serialize(group)

    def test_serialize_unstored_computer(self):
        """Test that you can't serialize an unstored node"""
        computer = orm.Computer('test_computer', 'test_host')

        with self.assertRaises(ValueError):
            serialize.serialize(computer)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for `verdi config`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.backends.tests.utils.configuration import with_temporary_config_instance
from aiida.cmdline.commands import cmd_verdi
from aiida.manage.configuration import get_config


class TestVerdiConfig(AiidaTestCase):
    """Tests for `verdi config`."""

    def setUp(self):
        self.cli_runner = CliRunner()

    @with_temporary_config_instance
    def test_config_option_set(self):
        """Test the `verdi config` command when setting an option."""
        config = get_config()

        option_name = 'daemon.timeout'
        option_values = [str(10), str(20)]

        for option_value in option_values:
            options = ['config', option_name, str(option_value)]
            result = self.cli_runner.invoke(cmd_verdi.verdi, options)

            self.assertClickSuccess(result)
            self.assertEqual(str(config.option_get(option_name, scope=config.current_profile.name)), option_value)

    @with_temporary_config_instance
    def test_config_option_get(self):
        """Test the `verdi config` command when getting an option."""
        option_name = 'daemon.timeout'
        option_value = str(30)

        options = ['config', option_name, option_value]
        result = self.cli_runner.invoke(cmd_verdi.verdi, options)
        self.assertClickSuccess(result)
        self.assertClickResultNoException(result)

        options = ['config', option_name]
        result = self.cli_runner.invoke(cmd_verdi.verdi, options)
        self.assertClickSuccess(result)
        self.assertIn(option_value, result.output.strip())

    @with_temporary_config_instance
    def test_config_option_unset(self):
        """Test the `verdi config` command when unsetting an option."""
        option_name = 'daemon.timeout'
        option_value = str(30)

        options = ['config', option_name, str(option_value)]
        result = self.cli_runner.invoke(cmd_verdi.verdi, options)
        self.assertClickSuccess(result)

        options = ['config', option_name]
        result = self.cli_runner.invoke(cmd_verdi.verdi, options)
        self.assertClickSuccess(result)
        self.assertIn(option_value, result.output.strip())

        options = ['config', option_name, '--unset']
        result = self.cli_runner.invoke(cmd_verdi.verdi, options)
        self.assertClickSuccess(result)
        self.assertIn('{} unset'.format(option_name), result.output.strip())

        options = ['config', option_name]
        result = self.cli_runner.invoke(cmd_verdi.verdi, options)
        self.assertClickSuccess(result)
        self.assertEqual(result.output, '')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import datetime
import importlib
import shutil
import sys
import tempfile

from dateutil.parser import parse

from aiida.backends.utils import is_dbenv_loaded, load_dbenv, BACKEND_SQLA, BACKEND_DJANGO
from aiida.backends.settings import BACKEND
from aiida.backends.testbase import AiidaTestCase
from aiida.common import utils
from aiida.manage.backup import backup_setup
from aiida.manage.backup import backup_utils
from aiida.orm import Data
from aiida.common import json


if not is_dbenv_loaded():
    load_dbenv()


class TestBackupScriptUnit(AiidaTestCase):


    _json_test_input_1 = '{"backup_length_threshold": 2, "periodicity": 2,' + \
        ' "oldest_object_backedup": "2014-07-18 13:54:53.688484+00:00", ' + \
        '"end_date_of_backup": null, "days_to_backup": null, "backup_dir": ' +\
        '"/scratch/aiida_user/backupScriptDest"}'

    _json_test_input_2 = '{"backup_length_threshold": 2, "periodicity": 2, ' +\
        '"oldest_object_backedup": "2014-07-18 13:54:53.688484+00:00", ' + \
        '"end_date_of_backup": null, "days_to_backup": null, "backup_dir": ' +\
        '"/scratch/aiida_user/backupScriptDest"}'

    _json_test_input_3 = '{"backup_length_threshold": 2, "periodicity": 2, ' +\
        '"oldest_object_backedup": "2014-07-18 13:54:53.688484+00:00", ' + \
        '"end_date_of_backup": null, "days_to_backup": 2, "backup_dir": ' + \
        '"/scratch/aiida_user/backupScriptDest"}'

    _json_test_input_4 = '{"backup_length_threshold": 2, "periodicity": 2, ' +\
        '"oldest_object_backedup": "2014-07-18 13:54:53.688484+00:00", ' + \
        '"end_date_of_backup": "2014-07-22 14:54:53.688484+00:00", ' + \
        '"days_to_backup": null, "backup_dir": ' + \
        '"/scratch/aiida_user/backupScriptDest"}'

    _json_test_input_5 = '{"backup_length_threshold": 2, "periodicity": 2, ' +\
        '"oldest_object_backedup": "2014-07-18 13:54:53.688484+00:00", ' + \
        '"end_date_of_backup": "2014-07-22 14:54:53.688484+00:00", ' + \
        '"days_to_backup": 2, "backup_dir": "/scratch/aiida_user/backup"}'

    _json_test_input_6 = '{"backup_length_threshold": 2, "periodicity": 2, ' +\
        '"oldest_object_backedup": "2014-07-18 13:54:53.688484", ' + \
        '"end_date_of_backup": "2014-07-22 14:54:53.688484", ' + \
        '"days_to_backup": null, ' \
        '"backup_dir": "/scratch/./aiida_user////backup//"}'

    def setUp(self):
        super(TestBackupScriptUnit, self).setUp()
        if not is_dbenv_loaded():
            load_dbenv()

        if BACKEND == BACKEND_SQLA:
            from aiida.manage.backup.backup_sqlalchemy import Backup
        elif BACKEND == BACKEND_DJANGO:
            from aiida.manage.backup.backup_django import Backup
        else:
            self.skipTest("Unknown backend")

        self._backup_setup_inst = Backup("", 2)

    def tearDown(self):
        super(TestBackupScriptUnit, self).tearDown()
        self._backup_setup_inst = None

    def test_loading_basic_params_from_file(self):
        """
        This method tests the correct loading of the basic _backup_setup_inst
        parameters from a JSON string.
        """
        backup_variables = json.loads(self._json_test_input_1)
        self._backup_setup_inst._ignore_backup_dir_existence_check = True
        self._backup_setup_inst._read_backup_info_from_dict(backup_variables)

        self.assertEqual(
            self._backup_setup_inst._oldest_object_bk,
            parse("2014-07-18 13:54:53.688484+00:00"),
            "Last _backup_setup_inst start date is not parsed correctly")

        # The destination directory of the _backup_setup_inst
        self.assertEqual(
            self._backup_setup_inst._backup_dir,
            "/scratch/aiida_user/backupScriptDest",
            "_backup_setup_inst destination directory not parsed correctly")

        self.assertEqual(
            self._backup_setup_inst._backup_length_threshold,
            datetime.timedelta(hours=2),
            "_backup_length_threshold not parsed correctly")

        self.assertEqual(
            self._backup_setup_inst._periodicity,
            2,
            "_periodicity not parsed correctly")

    def test_loading_backup_time_params_from_file_1(self):
        """
        This method tests that the _backup_setup_inst limits are correctly
        loaded from the JSON string and are correctly set.

        In the parsed JSON string, no _backup_setup_inst end limits are set
        """
        backup_variables = json.loads(self._json_test_input_2)
        self._backup_setup_inst._ignore_backup_dir_existence_check = True
        self._backup_setup_inst._read_backup_info_from_dict(backup_variables)

        self.assertEqual(
            self._backup_setup_inst._days_to_backup,
            None,
            "_days_to_backup should be None/null but it is not")

        self.assertEqual(
            self._backup_setup_inst._end_date_of_backup,
            None,
            "_end_date_of_backup should be None/null but it is not")

        self.assertEqual(
            self._backup_setup_inst._internal_end_date_of_backup,
            None,
            "_internal_end_date_of_backup should be None/null but it is not")

    def test_loading_backup_time_params_from_file_2(self):
        """
        This method tests that the _backup_setup_inst limits are correctly
        loaded from the JSON string and are correctly set.

        In the parsed JSON string, only the daysToBackup limit is set.
        """
        backup_variables = json.loads(self._json_test_input_3)
        self._backup_setup_inst._ignore_backup_dir_existence_check = True
        self._backup_setup_inst._read_backup_info_from_dict(backup_variables)

        self.assertEqual(
            self._backup_setup_inst._days_to_backup,
            2,
            "_days_to_backup should be 2 but it is not")

        self.assertEqual(
            self._backup_setup_inst._end_date_of_backup,
            None,
            "_end_date_of_backup should be None/null but it is not")

        self.assertEqual(
            self._backup_setup_inst._internal_end_date_of_backup,
            parse("2014-07-20 13:54:53.688484+00:00"),
            "_internal_end_date_of_backup is not the expected one")

    def test_loading_backup_time_params_from_file_3(self):
        """
        This method tests that the _backup_setup_inst limits are correctly
        loaded from the JSON string and are correctly set.

        In the parsed JSON string, only the endDateOfBackup limit is set.
        """
        backup_variables = json.loads(self._json_test_input_4)
        self._backup_setup_inst._ignore_backup_dir_existence_check = True
        self._backup_setup_inst._read_backup_info_from_dict(backup_variables)

        self.assertEqual(
            self._backup_setup_inst._days_to_backup,
            None,
            "_days_to_backup should be None/null but it is not")

        self.assertEqual(
            self._backup_setup_inst._end_date_of_backup,
            parse("2014-07-22 14:54:53.688484+00:00"),
            "_end_date_of_backup should be None/null but it is not")

        self.assertEqual(
            self._backup_setup_inst._internal_end_date_of_backup,
            parse("2014-07-22 14:54:53.688484+00:00"),
            "_internal_end_date_of_backup is not the expected one")

    def test_loading_backup_time_params_from_file_4(self):
        """
        This method tests that the _backup_setup_inst limits are correctly
        loaded from the JSON string and are correctly set.

        In the parsed JSON string, the endDateOfBackup & daysToBackuplimit
        are set which should lead to an exception.
        """
        from aiida.manage.backup.backup_base import BackupError

        backup_variables = json.loads(self._json_test_input_5)
        self._backup_setup_inst._ignore_backup_dir_existence_check = True
        # An exception should be raised because endDateOfBackup
        # & daysToBackuplimit have been defined in the same time.
        with self.assertRaises(BackupError):
            self._backup_setup_inst._read_backup_info_from_dict(backup_variables)

    def check_full_deserialization_serialization(self, input_string, backup_inst):
        input_variables = json.loads(input_string)
        backup_inst._ignore_backup_dir_existence_check = True
        backup_inst._read_backup_info_from_dict(input_variables)
        target_variables = backup_inst._dictionarize_backup_info()

        self.assertEqual(input_variables, target_variables,
                      "The test string {} did not succeed".format(
                          input_string) +
                      " the serialization deserialization test.\n" +
                      "Input variables: {}\n".format(input_variables) +
                      "Output variables: {}\n".format(target_variables))

    def test_full_deserialization_serialization_1(self):
        """
        This method tests the correct deserialization / serialization of the
        variables that should be stored in a file.
        """
        input_string = self._json_test_input_1
        backup_inst = self._backup_setup_inst

        self.check_full_deserialization_serialization(input_string, backup_inst)

    def test_full_deserialization_serialization_2(self):
        """
        This method tests the correct deserialization / serialization of the
        variables that should be stored in a file.
        """
        input_string = self._json_test_input_2
        backup_inst = self._backup_setup_inst

        self.check_full_deserialization_serialization(input_string, backup_inst)

    def test_full_deserialization_serialization_3(self):
        """
        This method tests the correct deserialization / serialization of the
        variables that should be stored in a file.
        """
        input_string = self._json_test_input_3
        backup_inst = self._backup_setup_inst

        self.check_full_deserialization_serialization(input_string, backup_inst)

    def test_full_deserialization_serialization_4(self):
        """
        This method tests the correct deserialization / serialization of the
        variables that should be stored in a file.
        """
        input_string = self._json_test_input_4
        backup_inst = self._backup_setup_inst

        self.check_full_deserialization_serialization(input_string, backup_inst)

    def test_timezone_addition_and_dir_correction(self):
        """
        This method tests if the timezone is added correctly to timestamps
        that don't have a timezone. Moreover, it checks if the given directory
        paths are normalized as expected.
        """

        backup_variables = json.loads(self._json_test_input_6)
        self._backup_setup_inst._ignore_backup_dir_existence_check = True
        self._backup_setup_inst._read_backup_info_from_dict(backup_variables)

        self.assertIsNotNone(
            self._backup_setup_inst._oldest_object_bk.tzinfo,
            "Timezone info should not be none (timestamp: {})."
            .format(self._backup_setup_inst._oldest_object_bk))

        self.assertIsNotNone(
            self._backup_setup_inst._end_date_of_backup.tzinfo,
            "Timezone info should not be none (timestamp: {})."
            .format(self._backup_setup_inst._end_date_of_backup))

        self.assertIsNotNone(
            self._backup_setup_inst._internal_end_date_of_backup.tzinfo,
            "Timezone info should not be none (timestamp: {})."
            .format(self._backup_setup_inst._internal_end_date_of_backup))

        # The destination directory of the _backup_setup_inst
        self.assertEqual(
            self._backup_setup_inst._backup_dir,
            "/scratch/aiida_user/backup",
            "_backup_setup_inst destination directory is "
            "not normalized as expected.")


class TestBackupScriptIntegration(AiidaTestCase):

    _aiida_rel_path = ".aiida"
    _backup_rel_path = "backup"
    _repo_rel_path = "repository"

    _bs_instance = backup_setup.BackupSetup()

    def test_integration(self):
        from aiida.common.utils import Capturing

        # Fill in the repository with data
        self.fill_repo()
        try:
            # Create a temp folder where the backup files will be placed
            # and the backup will be stored
            temp_folder = tempfile.mkdtemp()

            # Capture the sysout of the following command
            with Capturing():
                # Create the backup scripts
                backup_full_path = self.create_backup_scripts(temp_folder)

            # Put the backup folder in the path
            sys.path.append(backup_full_path)

            # Import the backup script - this action will also run it
            # It is assumed that the backup script ends with .py
            importlib.import_module(self._bs_instance._script_filename[:-3])

            # Check the backup
            from aiida import settings
            from filecmp import dircmp
            import os
            from aiida.common.utils import are_dir_trees_equal
            source_dir = os.path.join(settings.REPOSITORY_PATH,
                                      self._repo_rel_path)
            dest_dir = os.path.join(backup_full_path,
                                    self._bs_instance._file_backup_folder_rel,
                                    self._repo_rel_path)
            res, msg = are_dir_trees_equal(source_dir, dest_dir)
            self.assertTrue(res, "The backed-up repository has differences to the original one. " + str(msg)
                             + ". If the test fails, report it in issue #2134.")
        finally:
            shutil.rmtree(temp_folder, ignore_errors=True)

    def fill_repo(self):
        from aiida.orm import CalcJobNode, Data
        from aiida.plugins import CalculationFactory, DataFactory

        extra_name = self.__class__.__name__ + "/test_with_subclasses"
        resources = {'num_machines': 1, 'num_mpiprocs_per_machine': 1}

        Dict = DataFactory('dict')

        a1 = CalcJobNode(computer=self.computer)
        a1.set_option('resources', resources)
        a1.store()
        # To query only these nodes later
        a1.set_extra(extra_name, True)
        a3 = Data().store()
        a3.set_extra(extra_name, True)
        a4 = Dict(dict={'a': 'b'}).store()
        a4.set_extra(extra_name, True)
        a5 = Data().store()
        a5.set_extra(extra_name, True)
        # I don't set the extras, just to be sure that the filtering works
        # The filtering is needed because other tests will put stuff int he DB
        a6 = CalcJobNode(computer=self.computer)
        a6.set_option('resources', resources)
        a6.store()
        a7 = Data()
        a7.store()

    def create_backup_scripts(self, tmp_folder):
        backup_full_path = "{}/{}/{}/".format(tmp_folder, self._aiida_rel_path,
                                              self._backup_rel_path)
        # The predefined answers for the setup script
        ac = utils.ArrayCounter()
        answers = [backup_full_path,   # the backup folder path
                   "",                  # should the folder be created?
                   "",                  # destination folder of the backup
                   "",                  # should the folder be created?
                   "n",                 # print config explanation?
                   "",                  # configure the backup conf file now?
                   "", # start date of backup?
                   "",                  # is it correct?
                   "",                  # days to backup?
                   "",                  # is it correct?
                   "", # end date of backup
                   "",                  # is it correct?
                   "1",                 # periodicity
                   "",                  # is it correct?
                   "0",                 # threshold?
                   ""]                  # is it correct?
        backup_utils.input = lambda _: answers[ac.array_counter()]

        # Run the setup script
        self._bs_instance.run()

        return backup_full_path



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi status` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import sys
from enum import IntEnum
import click

from aiida.cmdline.commands.cmd_verdi import verdi
from ..utils.echo import ExitCode


class ServiceStatus(IntEnum):
    """Describe status of services for 'verdi status' command.
    """
    UP = 0  # pylint: disable=invalid-name
    ERROR = 1
    DOWN = 2


STATUS_SYMBOLS = {
    ServiceStatus.UP: {
        'color': 'green',
        'string': u'\u2713',
    },
    ServiceStatus.ERROR: {
        'color': 'red',
        'string': u'\u2717',
    },
    ServiceStatus.DOWN: {
        'color': 'red',
        'string': u'\u2717',
    },
}


@verdi.command('status')
def verdi_status():
    """Print status of AiiDA services."""
    # pylint: disable=broad-except,too-many-statements
    from aiida.manage.external.rmq import get_rmq_url
    from aiida.manage.external.postgres import Postgres
    from aiida.cmdline.utils.daemon import get_daemon_status
    from aiida.common.utils import Capturing, get_repository_folder
    from aiida.manage.manager import get_manager

    exit_code = ExitCode.SUCCESS
    manager = get_manager()

    # getting the profile
    try:
        profile = manager.get_profile()
        print_status(ServiceStatus.UP, 'profile', "On profile {}".format(profile.name))

    except Exception as exc:
        print_status(ServiceStatus.ERROR, 'profile', "Unable to read AiiDA profile")
        exit_code = ExitCode.CRITICAL
        sys.exit(exit_code)  # stop here - without a profile we cannot access anything

    # getting the repository
    try:
        repo_folder = get_repository_folder()
    except Exception as exc:
        print_status(ServiceStatus.ERROR, 'repository', "Error with repo folder", exception=exc)
        exit_code = ExitCode.CRITICAL

    print_status(ServiceStatus.UP, 'repository', repo_folder)

    # getting the postgres status
    try:
        with Capturing(capture_stderr=True):

            postgres = Postgres.from_profile(profile)
            pg_connected = postgres.try_connect()

        dbinfo = postgres.get_dbinfo()
        if pg_connected:
            print_status(ServiceStatus.UP, 'postgres', "Connected to {}@{}:{}".format(
                dbinfo['user'], dbinfo['host'], dbinfo['port']))
        else:
            print_status(ServiceStatus.DOWN, 'postgres', "Unable to connect to {}:{}".format(
                dbinfo['host'], dbinfo['port']))
            exit_code = ExitCode.CRITICAL

    except Exception as exc:
        pd_dict = profile.dictionary
        print_status(
            ServiceStatus.ERROR,
            'postgres',
            "Error connecting to {}:{}".format(pd_dict['AIIDADB_HOST'], pd_dict['AIIDADB_PORT']),
            exception=exc)
        exit_code = ExitCode.CRITICAL

    # getting the rmq status
    try:
        with Capturing(capture_stderr=True):
            comm = manager.create_communicator(with_orm=False)
            comm.stop()

        print_status(ServiceStatus.UP, 'rabbitmq', "Connected to {}".format(get_rmq_url()))

    except Exception as exc:
        print_status(ServiceStatus.ERROR, 'rabbitmq', "Unable to connect to rabbitmq", exception=exc)
        exit_code = ExitCode.CRITICAL

    # getting the daemon status
    try:
        client = manager.get_daemon_client()
        daemon_status = get_daemon_status(client)

        daemon_status = daemon_status.split("\n")[0]  # take only the first line
        if client.is_daemon_running:
            print_status(ServiceStatus.UP, 'daemon', daemon_status)
        else:
            print_status(ServiceStatus.DOWN, 'daemon', daemon_status)
            exit_code = ExitCode.CRITICAL

    except Exception as exc:
        print_status(ServiceStatus.ERROR, 'daemon', "Error getting daemon status", exception=exc)
        exit_code = ExitCode.CRITICAL

    # Note: click does not forward return values to the exit code, see
    # https://github.com/pallets/click/issues/747
    sys.exit(exit_code)


def print_status(status, service, msg="", exception=None):
    """Print status message.

    Includes colored indicator.

    :param status: a ServiceStatus code
    :param service: string for service name
    :param msg:  message string

    """
    symbol = STATUS_SYMBOLS[status]
    click.secho(u' {} '.format(symbol['string']), fg=symbol['color'], nl=False)
    click.secho('{:12s} {}'.format(service + ':', msg))
    if exception is not None:
        click.echo(exception, err=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import sys

from alembic import command as alembic_command
from aiida.backends import sqlalchemy as sa
from aiida.backends.sqlalchemy.utils import get_alembic_conf

# Available alembic commands
REVISION_CMD = 'revision'
CURRENT_CMD = 'current'
HISTORY_CMD = 'history'
UPGRADE_CMD = 'upgrade'
DOWNGRADE_CMD = 'downgrade'

AVAIL_AL_COMMANDS = [REVISION_CMD, CURRENT_CMD, HISTORY_CMD,
                     UPGRADE_CMD, DOWNGRADE_CMD]

if __name__ == "__main__":
    import argparse
    from aiida.backends.profile import load_profile
    from aiida.backends.sqlalchemy.utils import _load_dbenv_noschemacheck
    from aiida.backends.profile import BACKEND_SQLA
    from aiida.common.exceptions import InvalidOperation

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--aiida-profile', help='The AiiDA profile that you would like to use')
    parser.add_argument('--no-autogenerate', action='store_true', help='Only generate an empty revision template file '
        'but do not automatically generate the migration operations')

    subparsers = parser.add_subparsers(
        help='sub-command help', dest='command')

    parser_upg = subparsers.add_parser('upgrade', help='Upgrade to a later version')
    parser_upg.add_argument('revision', help='specify the revision')

    parser_dg = subparsers.add_parser('downgrade', help='Revert to a previous version')
    parser_dg.add_argument('revision', help='specify the revision')

    parser_hist = subparsers.add_parser('history', help='List changeset scripts in chronological order')
    parser_hist.add_argument('-r', '--rev-range', help='Output in verbose mode')
    parser_hist.add_argument('-v', '--verbose', action='store_true', help='Output in verbose mode')

    parser_cur = subparsers.add_parser('current', help='Display the current version for a database')
    parser_cur.add_argument('-v', '--verbose', action='store_true', help='Output in verbose mode')

    parser_rev = subparsers.add_parser(
        'revision', help='Create a new migration file')
    parser_rev.add_argument(
        'arguments', nargs=1, help='Migration message')

    args = parser.parse_args(sys.argv[1:])

    if args.command in AVAIL_AL_COMMANDS:
        # Use the default profile if not specified
        profile_name = args.aiida_profile

        # Perform the same loading procedure as the normal load_dbenv does
        from aiida.backends import settings
        settings.LOAD_DBENV_CALLED = True
        # We load the needed profile.
        # This is going to set global variables in settings, including
        # settings.BACKEND
        load_profile(profile=profile_name)
        if settings.BACKEND != BACKEND_SQLA:
            raise InvalidOperation("A SQLAlchemy (alembic) revision "
                                   "generation procedure is initiated "
                                   "but a different backend is used!")
        _load_dbenv_noschemacheck(profile=profile_name)

        # Get the requested alembic command from the available commands
        command = getattr(alembic_command, args.command)
        alembic_cfg = get_alembic_conf()

        with sa.engine.begin() as connection:
            alembic_cfg.attributes['connection'] = connection
            if command.__name__ == CURRENT_CMD:
                command(alembic_cfg, verbose=args.verbose)
            elif command.__name__ == HISTORY_CMD:
                command(alembic_cfg, rev_range=args.rev_range, verbose=args.verbose)
            elif command.__name__ == REVISION_CMD:
                command(alembic_cfg, message=args.arguments[0], autogenerate=not args.no_autogenerate)
            elif command.__name__ in [UPGRADE_CMD, DOWNGRADE_CMD]:
                command(alembic_cfg, revision=args.revision)
            else:
                raise ValueError('invalid command')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test for the `CalcJob` process sub class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.common import exceptions
from aiida.engine import launch, CalcJob, Process, CalcJobBuilder
from aiida.plugins import CalculationFactory

ArithmeticAddCalculation = CalculationFactory('arithmetic.add')  # pylint: disable=invalid-name


class TestCalcJob(AiidaTestCase):
    """Test for the `CalcJob` process sub class."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestCalcJob, cls).setUpClass(*args, **kwargs)
        cls.code = orm.Code(remote_computer_exec=(cls.computer, '/bin/true')).store()

    def setUp(self):
        super(TestCalcJob, self).setUp()
        self.assertIsNone(Process.current())

    def tearDown(self):
        super(TestCalcJob, self).tearDown()
        self.assertIsNone(Process.current())

    def test_run_base_class(self):
        """Verify that it is impossible to run, submit or instantiate a base `CalcJob` class."""
        with self.assertRaises(exceptions.InvalidOperation):
            CalcJob()

        with self.assertRaises(exceptions.InvalidOperation):
            launch.run(CalcJob)

        with self.assertRaises(exceptions.InvalidOperation):
            launch.run_get_node(CalcJob)

        with self.assertRaises(exceptions.InvalidOperation):
            launch.run_get_pk(CalcJob)

        with self.assertRaises(exceptions.InvalidOperation):
            launch.submit(CalcJob)

    def test_define_not_calling_super(self):
        """A `CalcJob` that does not call super in `define` classmethod should raise."""

        class IncompleteDefineCalcJob(CalcJob):
            """Test class with incomplete define method"""

            @classmethod
            def define(cls, spec):
                pass

            def prepare_for_submission(self, folder):
                pass

        with self.assertRaises(AssertionError):
            launch.run(IncompleteDefineCalcJob)

    def test_invalid_options_type(self):
        """Verify that passing an invalid type to `metadata.options` raises a `TypeError`."""

        class SimpleCalcJob(CalcJob):
            """Simple `CalcJob` implementation"""

            @classmethod
            def define(cls, spec):
                super(SimpleCalcJob, cls).define(spec)

            def prepare_for_submission(self, folder):
                pass

        # The `metadata.options` input expects a plain dict and not a node `Dict`
        with self.assertRaises(TypeError):
            launch.run(SimpleCalcJob, code=self.code, metadata={'options': orm.Dict(dict={'a': 1})})

    def test_return_builder(self):
        """ verify that `CalcJob.get_builder` returns CalcJobBuilder"""
        builder = CalcJob.get_builder()
        self.assertIsInstance(builder, CalcJobBuilder)

    def test_invalid_parser_name(self):
        """Passing an invalid parser name should already stop during input validation."""
        inputs = {
            'code': self.code,
            'x': orm.Int(1),
            'y': orm.Int(2),
            'metadata': {
                'options': {
                    'resources': {
                        'num_machines': 1,
                        'num_mpiprocs_per_machine': 1
                    },
                    'parser_name': 'invalid_parser'
                }
            }
        }

        with self.assertRaises(exceptions.ValidationError):
            ArithmeticAddCalculation(inputs=inputs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for the `AuthInfo` ORM class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.common import exceptions
from aiida.plugins import TransportFactory
from aiida.manage.manager import get_manager
from . import entities
from . import users

__all__ = ('AuthInfo',)


class AuthInfo(entities.Entity):
    """ORM class that models the authorization information that allows a `User` to connect to a `Computer`."""

    class Collection(entities.Collection):
        """The collection of `AuthInfo` entries."""

        def delete(self, pk):
            """Delete an entry from the collection.

            :param pk: the pk of the entry to delete
            """
            self._backend.authinfos.delete(pk)

    PROPERTY_WORKDIR = 'workdir'

    def __init__(self, computer, user, backend=None):
        """Create an `AuthInfo` instance for the given computer and user.

        :param computer: a `Computer` instance
        :param user: a `User` instance
        :return: :class:`aiida.orm.authinfos.AuthInfo`
        """
        backend = backend or get_manager().get_backend()
        model = backend.authinfos.create(computer=computer.backend_entity, user=user.backend_entity)
        super(AuthInfo, self).__init__(model)

    def __str__(self):
        if self.enabled:
            return 'AuthInfo for {} on {}'.format(self.user.email, self.computer.name)

        return 'AuthInfo for {} on {} [DISABLED]'.format(self.user.email, self.computer.name)

    @property
    def enabled(self):
        """Return whether this instance is enabled.

        :return: boolean, True if enabled, False otherwise
        """
        return self._backend_entity.enabled

    @enabled.setter
    def enabled(self, enabled):
        """Set the enabled state

        :param enabled: boolean, True to enable the instance, False to disable it
        """
        self._backend_entity.enabled = enabled

    @property
    def computer(self):
        """Return the computer associated with this instance.

        :return: :class:`aiida.orm.computers.Computer`
        """
        from . import computers  # pylint: disable=cyclic-import
        return computers.Computer.from_backend_entity(self._backend_entity.computer)

    @property
    def user(self):
        """Return the user associated with this instance.

        :return: :class:`aiida.orm.users.User`
        """
        return users.User.from_backend_entity(self._backend_entity.user)

    def get_auth_params(self):
        """Return the dictionary of authentication parameters

        :return: a dictionary with authentication parameters
        """
        return self._backend_entity.get_auth_params()

    def set_auth_params(self, auth_params):
        """Set the dictionary of authentication parameters

        :param auth_params: a dictionary with authentication parameters
        """
        self._backend_entity.set_auth_params(auth_params)

    def get_metadata(self):
        """Return the dictionary of metadata

        :return: a dictionary with metadata
        """
        return self._backend_entity.get_metadata()

    def set_metadata(self, metadata):
        """Set the dictionary of metadata

        :param metadata: a dictionary with metadata
        """
        self._backend_entity.set_metadata(metadata)

    def get_workdir(self):
        """Return the working directory.

        If no explicit work directory is set for this instance, the working directory of the computer will be returned.

        :return: the working directory
        :rtype: str
        """
        try:
            return self.get_metadata()[self.PROPERTY_WORKDIR]
        except KeyError:
            return self.computer.get_workdir()

    def get_transport(self):
        """Return a fully configured transport that can be used to connect to the computer set for this instance.

        :return: :class:`aiida.transports.Transport`
        """
        computer = self.computer
        transport_type = computer.get_transport_type()
        transport_params = computer.get_transport_params()

        try:
            transport_class = TransportFactory(transport_type)
        except exceptions.EntryPointError as exception:
            raise exceptions.ConfigurationError('transport type `{}` could not be loaded: {}'.format(
                transport_type, exception))

        parameters = dict(list(transport_params.items()) + list(self.get_auth_params().items()))

        return transport_class(machine=computer.hostname, **parameters)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the `verdi group` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.common import exceptions
from aiida.cmdline.commands.cmd_group import (group_list, group_create, group_delete, group_relabel, group_description,
                                              group_add_nodes, group_remove_nodes, group_show, group_copy)


class TestVerdiGroup(AiidaTestCase):
    """Tests for the `verdi group` command."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestVerdiGroup, cls).setUpClass(*args, **kwargs)
        for group in ['dummygroup1', 'dummygroup2', 'dummygroup3', 'dummygroup4']:
            orm.Group(label=group).store()

    def setUp(self):
        """Create runner object to run tests."""
        from click.testing import CliRunner
        self.cli_runner = CliRunner()

    def test_help(self):
        """Tests help text for all group sub commands."""
        options = ['--help']

        # verdi group list
        result = self.cli_runner.invoke(group_list, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Usage', result.output)

        # verdi group create
        result = self.cli_runner.invoke(group_create, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Usage', result.output)

        # verdi group delete
        result = self.cli_runner.invoke(group_delete, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Usage', result.output)

        # verdi group relabel
        result = self.cli_runner.invoke(group_relabel, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Usage', result.output)

        # verdi group description
        result = self.cli_runner.invoke(group_description, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Usage', result.output)

        # verdi group addnodes
        result = self.cli_runner.invoke(group_add_nodes, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Usage', result.output)

        # verdi group removenodes
        result = self.cli_runner.invoke(group_remove_nodes, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Usage', result.output)

        # verdi group show
        result = self.cli_runner.invoke(group_show, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Usage', result.output)

        # verdi group copy
        result = self.cli_runner.invoke(group_copy, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Usage', result.output)

    def test_create(self):
        """Test `verdi group create` command."""
        result = self.cli_runner.invoke(group_create, ['dummygroup5'])
        self.assertClickResultNoException(result)

        # check if newly added group in present in list
        result = self.cli_runner.invoke(group_list)
        self.assertClickResultNoException(result)

        self.assertIn('dummygroup5', result.output)

    def test_list(self):
        """Test `verdi group list` command."""
        result = self.cli_runner.invoke(group_list)
        self.assertClickResultNoException(result)

        for grp in ['dummygroup1', 'dummygroup2']:
            self.assertIn(grp, result.output)

    def test_copy(self):
        """Test `verdi group copy` command."""
        result = self.cli_runner.invoke(group_copy, ['dummygroup1', 'dummygroup2'])
        self.assertClickResultNoException(result)

        self.assertIn('Success', result.output)

    def test_delete(self):
        """Test `verdi group delete` command."""
        group_01 = orm.Group(label='group_test_delete_01').store()
        group_02 = orm.Group(label='group_test_delete_02').store()

        result = self.cli_runner.invoke(group_delete, ['--force', 'group_test_delete_01'])
        self.assertClickResultNoException(result)

        # Verify that removed group is not present in list
        result = self.cli_runner.invoke(group_list)
        self.assertClickResultNoException(result)
        self.assertNotIn('group_test_delete_01', result.output)

        node_01 = orm.CalculationNode().store()
        node_02 = orm.CalculationNode().store()

        # Add some nodes and then use `verdi group delete --clear` to delete a node even when it contains nodes
        group = orm.load_group(label='group_test_delete_02')
        group.add_nodes([node_01, node_02])
        self.assertEqual(group.count(), 2)

        # Calling delete on a group without the `--clear` option should raise
        result = self.cli_runner.invoke(group_delete, ['--force', 'group_test_delete_02'])
        self.assertIsNotNone(result.exception, result.output)

        # With `--clear` option should delete group and nodes
        result = self.cli_runner.invoke(group_delete, ['--force', '--clear', 'group_test_delete_02'])
        self.assertClickResultNoException(result)

        with self.assertRaises(exceptions.NotExistent):
            group = orm.load_group(label='group_test_delete_02')

    def test_show(self):
        """Test `verdi group show` command."""
        result = self.cli_runner.invoke(group_show, ['dummygroup1'])
        self.assertClickResultNoException(result)

        for grpline in [
            'Group label', 'dummygroup1', 'Group type_string', 'user', 'Group description', '<no description>'
        ]:
            self.assertIn(grpline, result.output)

    def test_description(self):
        """Test `verdi group description` command."""
        description = 'It is a new description'
        group = orm.load_group(label='dummygroup2')
        self.assertNotEqual(group.description, description)

        # Change the description of the group
        result = self.cli_runner.invoke(group_description, [group.label, description])
        self.assertClickResultNoException(result)
        self.assertEqual(group.description, description)

        # When no description argument is passed the command should just echo the current description
        result = self.cli_runner.invoke(group_description, [group.label])
        self.assertClickResultNoException(result)
        self.assertIn(description, result.output)

    def test_relabel(self):
        """Test `verdi group relabel` command."""
        result = self.cli_runner.invoke(group_relabel, ['dummygroup4', 'relabeled_group'])
        self.assertIsNone(result.exception, result.output)

        # check if group list command shows changed group name
        result = self.cli_runner.invoke(group_list)
        self.assertClickResultNoException(result)
        self.assertNotIn('dummygroup4', result.output)
        self.assertIn('relabeled_group', result.output)

    def test_add_remove_nodes(self):
        """Test `verdi group remove-nodes` command."""
        node_01 = orm.CalculationNode().store()
        node_02 = orm.CalculationNode().store()
        node_03 = orm.CalculationNode().store()

        result = self.cli_runner.invoke(group_add_nodes, ['--force', '--group=dummygroup1', node_01.uuid])
        self.assertClickResultNoException(result)

        # Check if node is added in group using group show command
        result = self.cli_runner.invoke(group_show, ['dummygroup1'])
        self.assertClickResultNoException(result)
        self.assertIn('CalculationNode', result.output)
        self.assertIn(str(node_01.pk), result.output)

        # Remove same node
        result = self.cli_runner.invoke(group_remove_nodes, ['--force', '--group=dummygroup1', node_01.uuid])
        self.assertIsNone(result.exception, result.output)

        # Check if node is added in group using group show command
        result = self.cli_runner.invoke(group_show, ['-r', 'dummygroup1'])
        self.assertClickResultNoException(result)
        self.assertNotIn('CalculationNode', result.output)
        self.assertNotIn(str(node_01.pk), result.output)

        # Add all three nodes and then use `verdi group remove-nodes --clear` to remove them all
        group = orm.load_group(label='dummygroup1')
        group.add_nodes([node_01, node_02, node_03])
        self.assertEqual(group.count(), 3)

        result = self.cli_runner.invoke(group_remove_nodes, ['--force', '--clear', '--group=dummygroup1'])
        self.assertClickResultNoException(result)
        self.assertEqual(group.count(), 0)

    def test_copy_existing_group(self):
        """Test user is prompted to continue if destination group exists and is not empty"""
        source_label = "source_copy_existing_group"
        dest_label = "dest_copy_existing_group"

        # Create source group with nodes
        calc_s1 = orm.CalculationNode().store()
        calc_s2 = orm.CalculationNode().store()
        nodes_source_group = {str(node.uuid) for node in [calc_s1, calc_s2]}
        source_group = orm.Group(label=source_label).store()
        source_group.add_nodes([calc_s1, calc_s2])

        # Copy using `verdi group copy` - making sure all is successful
        options = [source_label, dest_label]
        result = self.cli_runner.invoke(group_copy, options)
        self.assertClickResultNoException(result)
        self.assertIn(
            "Success: Nodes copied from group<{}> to group<{}>".format(source_label, dest_label),
            result.output,
            result.exception)

        # Check destination group exists with source group's nodes
        dest_group = orm.load_group(label=dest_label)
        self.assertEqual(dest_group.count(), 2)
        nodes_dest_group = {str(node.uuid) for node in dest_group.nodes}
        self.assertSetEqual(nodes_source_group, nodes_dest_group)

        # Copy again, making sure an abort error is raised, since no user input can be made and default is abort
        result = self.cli_runner.invoke(group_copy, options)
        self.assertIsNotNone(result.exception, result.output)
        self.assertIn(
            "Warning: Destination group<{}> already exists and is not empty.".format(dest_label),
            result.output,
            result.exception)

        # Check destination group is unchanged
        dest_group = orm.load_group(label=dest_label)
        self.assertEqual(dest_group.count(), 2)
        nodes_dest_group = {str(node.uuid) for node in dest_group.nodes}
        self.assertSetEqual(nodes_source_group, nodes_dest_group)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the `Dict` class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.testbase import AiidaTestCase
from aiida.orm import Dict


class TestDict(AiidaTestCase):
    """Test for the `Dict` class."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestDict, cls).setUpClass(*args, **kwargs)
        cls.dictionary = {'value': 1, 'nested': {'dict': 'ionary'}}
        cls.node = Dict(dict=cls.dictionary)

    def test_keys(self):
        """Test the `keys` method."""
        self.assertEqual(sorted(self.node.keys()), sorted(self.dictionary.keys()))

    def test_get_dict(self):
        """Test the `get_dict` method."""
        self.assertEqual(self.node.get_dict(), self.dictionary)

    def test_dict_property(self):
        """Test the `dict` property."""
        self.assertEqual(self.node.dict.value, self.dictionary['value'])
        self.assertEqual(self.node.dict.nested, self.dictionary['nested'])

    def test_get_item(self):
        """Test the `__getitem__` method."""
        self.assertEqual(self.node['value'], self.dictionary['value'])
        self.assertEqual(self.node['nested'], self.dictionary['nested'])



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import operator

from six.moves import range, zip

from aiida.backends.testbase import AiidaTestCase
from aiida.common.exceptions import ModificationNotAllowed
from aiida.orm import load_node, List, Bool, Float, Int, Str, NumericType
from aiida.orm.nodes.data.bool import get_true_node, get_false_node


class TestList(AiidaTestCase):

    def test_creation(self):
        node = List()
        self.assertEqual(len(node), 0)
        with self.assertRaises(IndexError):
            node[0]

    def test_append(self):
        def do_checks(node):
            self.assertEqual(len(node), 1)
            self.assertEqual(node[0], 4)

        node = List()
        node.append(4)
        do_checks(node)

        # Try the same after storing
        node = List()
        node.append(4)
        node.store()
        do_checks(node)

    def test_extend(self):
        lst = [1, 2, 3]

        def do_checks(node):
            self.assertEqual(len(node), len(lst))
            # Do an element wise comparison
            for x, y in zip(lst, node):
                self.assertEqual(x, y)

        node = List()
        node.extend(lst)
        do_checks(node)
        # Further extend
        node.extend(lst)
        self.assertEqual(len(node), len(lst) * 2)

        # Do an element wise comparison
        for i in range(len(lst)):
            self.assertEqual(lst[i], node[i])
            self.assertEqual(lst[i], node[i % len(lst)])

        # Now try after storing
        node = List()
        node.extend(lst)
        node.store()
        do_checks(node)

    def test_mutability(self):
        node = List()
        node.append(5)
        node.store()

        # Test all mutable calls are now disallowed
        with self.assertRaises(ModificationNotAllowed):
            node.append(5)
        with self.assertRaises(ModificationNotAllowed):
            node.extend([5])
        with self.assertRaises(ModificationNotAllowed):
            node.insert(0, 2)
        with self.assertRaises(ModificationNotAllowed):
            node.remove(0)
        with self.assertRaises(ModificationNotAllowed):
            node.pop()
        with self.assertRaises(ModificationNotAllowed):
            node.sort()
        with self.assertRaises(ModificationNotAllowed):
            node.reverse()

    def test_store_load(self):
        node = List(list=[1, 2, 3])
        node.store()

        node_loaded = load_node(node.pk)
        assert node.get_list() == node_loaded.get_list()


class TestFloat(AiidaTestCase):

    def setUp(self):
        super(TestFloat, self).setUp()
        self.value = Float()
        self.all_types = [Int, Float, Bool, Str]

    def test_create(self):
        a = Float()
        # Check that initial value is zero
        self.assertAlmostEqual(a.value, 0.0)

        f = Float(6.0)
        self.assertAlmostEqual(f.value, 6.)
        self.assertAlmostEqual(f, Float(6.0))

        i = Int()
        self.assertAlmostEqual(i.value, 0)
        i = Int(6)
        self.assertAlmostEqual(i.value, 6)
        self.assertAlmostEqual(f, i)

        b = Bool()
        self.assertAlmostEqual(b.value, False)
        b = Bool(False)
        self.assertAlmostEqual(b.value, False)
        self.assertAlmostEqual(b.value, get_false_node())
        b = Bool(True)
        self.assertAlmostEqual(b.value, True)
        self.assertAlmostEqual(b.value, get_true_node())

        s = Str()
        self.assertAlmostEqual(s.value, "")
        s = Str('Hello')
        self.assertAlmostEqual(s.value, 'Hello')

    def test_load(self):
        for t in self.all_types:
            node = t()
            node.store()
            loaded = load_node(node.pk)
            self.assertAlmostEqual(node, loaded)

    def test_add(self):
        a = Float(4)
        b = Float(5)
        # Check adding two db Floats
        res = a + b
        self.assertIsInstance(res, NumericType)
        self.assertAlmostEqual(res, 9.0)

        # Check adding db Float and native (both ways)
        res = a + 5.0
        self.assertIsInstance(res, NumericType)
        self.assertAlmostEqual(res, 9.0)

        res = 5.0 + a
        self.assertIsInstance(res, NumericType)
        self.assertAlmostEqual(res, 9.0)

        # Inplace
        a = Float(4)
        a += b
        self.assertAlmostEqual(a, 9.0)

        a = Float(4)
        a += 5
        self.assertAlmostEqual(a, 9.0)

    def test_mul(self):
        a = Float(4)
        b = Float(5)
        # Check adding two db Floats
        res = a * b
        self.assertIsInstance(res, NumericType)
        self.assertAlmostEqual(res, 20.0)

        # Check adding db Float and native (both ways)
        res = a * 5.0
        self.assertIsInstance(res, NumericType)
        self.assertAlmostEqual(res, 20)

        res = 5.0 * a
        self.assertIsInstance(res, NumericType)
        self.assertAlmostEqual(res, 20.0)

        # Inplace
        a = Float(4)
        a *= b
        self.assertAlmostEqual(a, 20)

        a = Float(4)
        a *= 5
        self.assertAlmostEqual(a, 20)

    def test_power(self):
        a = Float(4)
        b = Float(2)

        res = a ** b
        self.assertAlmostEqual(res.value, 16.)

    def test_division(self):
        """Test the normal division operator."""
        a = Float(3)
        b = Float(2)

        self.assertAlmostEqual(a / b, 1.5)
        self.assertIsInstance(a / b, Float)

    def test_division_integer(self):
        """Test the integer division operator."""
        a = Float(3)
        b = Float(2)

        self.assertAlmostEqual(a // b, 1.0)
        self.assertIsInstance(a // b, Float)

    def test_modulo(self):
        a = Float(12.0)
        b = Float(10.0)

        self.assertAlmostEqual(a % b, 2.0)
        self.assertIsInstance(a % b, NumericType)
        self.assertAlmostEqual(a % 10.0, 2.0)
        self.assertIsInstance(a % 10.0, NumericType)
        self.assertAlmostEqual(12.0 % b, 2.0)
        self.assertIsInstance(12.0 % b, NumericType)


class TestFloatIntMix(AiidaTestCase):

    def test_operator(self):
        a = Float(2.2)
        b = Int(3)

        for op in [operator.add, operator.mul, operator.pow, operator.lt, operator.le, operator.gt, operator.ge, operator.iadd, operator.imul]:
            for x, y in [(a, b), (b, a)]:
                c = op(x, y)
                c_val = op(x.value, y.value)
                self.assertEqual(c._type, type(c_val))
                self.assertEqual(c, op(x.value, y.value))


class TestInt(AiidaTestCase):

    def test_division(self):
        """Test the normal division operator."""
        a = Int(3)
        b = Int(2)

        self.assertAlmostEqual(a / b, 1.5)
        self.assertIsInstance(a / b, Float)

    def test_division_integer(self):
        """Test the integer division operator."""
        a = Int(3)
        b = Int(2)

        self.assertAlmostEqual(a // b, 1)
        self.assertIsInstance(a // b, Int)

    def test_modulo(self):
        a = Int(12)
        b = Int(10)

        self.assertEqual(a % b, 2)
        self.assertIsInstance(a % b, NumericType)
        self.assertEqual(a % 10, 2)
        self.assertIsInstance(a % 10, NumericType)
        self.assertEqual(12 % b, 2)
        self.assertIsInstance(12 % b, NumericType)



#!/usr/bin/env runaiida
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This WorkChain example is a very contrived implementation of the infamous FizzBuzz problem, that serves to illustrate
the various logical blocks that one can incorporate into the outline of the workchain's spec.
"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from aiida.engine import WorkChain, run, while_, if_
from aiida.orm import Int


class OutlineWorkChain(WorkChain):

    @classmethod
    def define(cls, spec):
        super(OutlineWorkChain, cls).define(spec)
        spec.input('a', valid_type=Int)
        spec.outline(
            cls.setup,
            while_(cls.not_finished)(
                if_(cls.if_multiple_of_three_and_five)(
                    cls.report_fizz_buzz
                ).elif_(cls.if_multiple_of_five)(
                    cls.report_buzz
                ).elif_(cls.if_multiple_of_three)(
                    cls.report_fizz
                ).else_(
                    cls.report_number
                ),
                cls.decrement
            )
        )

    def setup(self):
        self.ctx.counter = abs(self.inputs.a.value)

    def not_finished(self):
        return self.ctx.counter > 0

    def if_multiple_of_three_and_five(self):
        return (self.ctx.counter % 3 == 0 and self.ctx.counter % 5 == 0)

    def if_multiple_of_five(self):
        return self.ctx.counter % 5 == 0

    def if_multiple_of_three(self):
        return self.ctx.counter % 3 == 0

    def report_fizz_buzz(self):
        print('FizzBuzz')

    def report_fizz(self):
        print('Fizz')

    def report_buzz(self):
        print('Buzz')

    def report_number(self):
        print(self.ctx.counter)

    def decrement(self):
        self.ctx.counter -= 1


def main():
    run(OutlineWorkChain, a=Int(16))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Function that starts a daemon runner."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import logging
import signal

from aiida.common.log import configure_logging
from aiida.engine.daemon.client import get_daemon_client
from aiida.manage.manager import get_manager

LOGGER = logging.getLogger(__name__)


def start_daemon():
    """
    Start a daemon runner for the currently configured profile
    """
    daemon_client = get_daemon_client()
    configure_logging(daemon=True, daemon_log_file=daemon_client.daemon_log_file)

    try:
        runner = get_manager().create_daemon_runner()
    except Exception as exception:
        LOGGER.exception('daemon runner failed to start')
        raise

    def shutdown_daemon(_num, _frame):
        LOGGER.info('Received signal to shut down the daemon runner')
        runner.close()

    signal.signal(signal.SIGINT, shutdown_daemon)
    signal.signal(signal.SIGTERM, shutdown_daemon)

    LOGGER.info('Starting a daemon runner')

    try:
        runner.start()
    except SystemError as exception:
        LOGGER.info('Received a SystemError: %s', exception)
        runner.close()

    LOGGER.info('Daemon runner stopped')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
"""Generic functions to verify the integrity of the database and optionally apply patches to fix problems."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# These are all the entry points from the `aiida.calculations` category as registered with the AiiDA registry
# on Tuesday December 4 at 13:00:00 UTC
registered_calculation_entry_points = [
    'ase.ase = aiida_ase.calculations.ase:AseCalculation',
    'castep.bs = aiida_castep.calculations.castep:CastepBSCalculation',
    'castep.castep = aiida_castep.calculations.castep:CastepCalculation',
    'castep.pot1d = aiida_castep.calculations.castep:Pot1dCalculation',
    'castep.spec = aiida_castep.calculations.castep:CastepSpectralCalculation',
    'castep.ts = aiida_castep.calculations.castep:CastepTSCalculation',
    'codtools.cifcellcontents = aiida_codtools.calculations.cifcellcontents:CifcellcontentsCalculation',
    'codtools.cifcodcheck = aiida_codtools.calculations.cifcodcheck:CifcodcheckCalculation',
    'codtools.cifcoddeposit = aiida_codtools.calculations.cifcoddeposit:CifcoddepositCalculation',
    'codtools.cifcodnumbers = aiida_codtools.calculations.cifcodnumbers:CifcodnumbersCalculation',
    'codtools.ciffilter = aiida_codtools.calculations.ciffilter:CiffilterCalculation',
    'codtools.cifsplitprimitive = aiida_codtools.calculations.cifsplitprimitive:CifsplitprimitiveCalculation',
    'cp2k = aiida_cp2k.calculations:Cp2kCalculation',
    'crystal17.basic = aiida_crystal17.calculations.cry_basic:CryBasicCalculation',
    'crystal17.immigrant = aiida_crystal17.calculations.cry_main_immigrant:CryMainImmigrantCalculation',
    'crystal17.main = aiida_crystal17.calculations.cry_main:CryMainCalculation',
    'ddec = aiida_ddec.calculations:DdecCalculation', 'diff = aiida_diff.calculations:DiffCalculation',
    'dynaphopy = aiida_lammps.calculations.dynaphopy: DynaphopyCalculation',
    'gollum.gollum = aiida_gollum.calculations.gollum:GollumCalculation',
    'gudhi.rdm = aiida_gudhi.calculations.rips:RipsDistanceMatrixCalculation',
    'kkr.kkr = aiida_kkr.calculations.kkr:KkrCalculation',
    'kkr.kkrimp = aiida_kkr.calculations.kkrimp:KkrimpCalculation',
    'kkr.kkrimporter = aiida_kkr.calculations.kkrimporter:KkrImporterCalculation',
    'kkr.voro = aiida_kkr.calculations.voro:VoronoiCalculation',
    'lammps.combinate = aiida_lammps.calculations.lammps.combinate:CombinateCalculation',
    'lammps.force = aiida_lammps.calculations.lammps.force:ForceCalculation',
    'lammps.md = aiida_lammps.calculations.lammps.md:MdCalculation',
    'lammps.optimize = aiida_lammps.calculations.lammps.optimize:OptimizeCalculation',
    'nwchem.basic = aiida_nwchem.calculations.basic:BasicCalculation',
    'nwchem.pymatgen = aiida_nwchem.calculations.nwcpymatgen:NwcpymatgenCalculation',
    'phonopy.phono3py = aiida_phonopy.calculations.phonopy.phono3py: Phono3pyCalculation',
    'phonopy.phonopy = aiida_phonopy.calculations.phonopy.phonopy: PhonopyCalculation',
    'phtools.dmatrix = aiida_phtools.calculations.distance_matrix:DistanceMatrixCalculation',
    'phtools.surface = aiida_phtools.calculations.pore_surface:PoreSurfaceCalculation',
    'qeq.eqeq = aiida_qeq.calculations.eqeq:EQeqCalculation', 'qeq.qeq = aiida_qeq.calculations.qeq:QeqCalculation',
    'quantumespresso.cp = aiida_quantumespresso.calculations.cp:CpCalculation',
    'quantumespresso.dos = aiida_quantumespresso.calculations.dos:DosCalculation',
    'quantumespresso.hp = aiida_quantumespresso_hp.calculations.hp:HpCalculation',
    'quantumespresso.matdyn = aiida_quantumespresso.calculations.matdyn:MatdynCalculation',
    'quantumespresso.namelists = aiida_quantumespresso.calculations.namelists:NamelistsCalculation',
    'quantumespresso.neb = aiida_quantumespresso.calculations.neb:NebCalculation',
    'quantumespresso.ph = aiida_quantumespresso.calculations.ph:PhCalculation',
    'quantumespresso.pp = aiida_quantumespresso.calculations.pp:PpCalculation',
    'quantumespresso.projwfc = aiida_quantumespresso.calculations.projwfc:ProjwfcCalculation',
    'quantumespresso.pw = aiida_quantumespresso.calculations.pw:PwCalculation',
    'quantumespresso.pw2wannier90 = aiida_quantumespresso.calculations.pw2wannier90:Pw2wannier90Calculation',
    'quantumespresso.pwimmigrant = aiida_quantumespresso.calculations.pwimmigrant:PwimmigrantCalculation',
    'quantumespresso.q2r = aiida_quantumespresso.calculations.q2r:Q2rCalculation',
    'raspa = aiida_raspa.calculations:RaspaCalculation',
    'siesta.siesta = aiida_siesta.calculations.siesta:SiestaCalculation',
    'siesta.stm = aiida_siesta.calculations.stm:STMCalculation', 'vasp.vasp = aiida_vasp.calcs.vasp:VaspCalculation',
    'vasp.vasp2w90 = aiida_vasp.calcs.vasp2w90:Vasp2w90Calculation',
    'wannier90.wannier90 = aiida_wannier90.calculations:Wannier90Calculation',
    'yambo.yambo =  aiida_yambo.calculations.gw:YamboCalculation',
    'zeopp.network = aiida_zeopp.calculations.network:NetworkCalculation'
]


def infer_calculation_entry_point(type_strings):
    """Try to infer a calculation entry point name for all the calculation type strings that are found in the database.

    Before the plugin system was introduced, the `type` column of the node table was a string based on the base node
    type with the module path and class name appended. For example, for the `PwCalculation` class, which was a sub
    class of `JobCalculation`, would get `calculation.job.quantumespresso.pw.PwCalculation.` as its `type` string.
    At this point, the `JobCalculation` also still fullfilled the role of both the `Process` class as well as the
    `Node` class. In the migration for `v1.0.0`, this had to be migrated, where the `type` became that of the actual
    node i.e. `node.process.calculation.calcjob.CalcJobNode.` which would lose the information of which actual sub class
    it represented. This information should be stored in the `process_type` column, where the value is the name of the
    entry point of that calculation class.

    This function will, for a given set of calculation type strings of pre v1.0.0, try to map them on the known entry
    points for the calculation category. This is the union of those entry points registered at the AiiDA registry (see
    the mapping above) and those available in the environment in which this function is ran.

    If a type string cannot be mapped onto an entry point name, a fallback `process_type` string will be generated
    which is based on part of the old `type` string. For example, `calculation.job.unknown.UnknownCalculation.` would
    get the process type string `~unknown.UnknownCalculation`.

    The function will return a mapping of type strings onto their inferred process type strings.

    :param type_strings: a set of type strings whose entry point is to be inferred
    :return: a mapping of current node type string to the inferred entry point name
    """
    from reentry.entrypoint import EntryPoint
    from aiida.plugins.entry_point import get_entry_points

    prefix_calc_job = 'calculation.job.'
    entry_point_group = 'aiida.calculations'

    entry_point_names = []
    mapping_node_type_to_entry_point = {}

    # Build the list of known entry points, joining those present in the environment with the hard-coded set from taken
    # from the aiida-registry. Note that if entry points with the same name are found in both sets, the entry point
    # from the local environment is kept as leading.
    entry_points_local = get_entry_points(group=entry_point_group)
    entry_points_registry = [EntryPoint.parse(entry_point) for entry_point in registered_calculation_entry_points]

    entry_points = entry_points_local
    entry_point_names = [entry_point.name for entry_point in entry_points]

    for entry_point in entry_points_registry:
        if entry_point.name not in entry_point_names:
            entry_point_names.append(entry_point.name)

    for type_string in type_strings:

        # If it does not start with the calculation job prefix, it cannot possibly reference a calculation plugin
        if not type_string.startswith(prefix_calc_job):
            continue

        plugin_string = type_string[len(prefix_calc_job):]
        plugin_parts = [part for part in plugin_string.split('.') if part]
        plugin_class = plugin_parts.pop()
        inferred_entry_point_name = '.'.join(plugin_parts)

        if inferred_entry_point_name in entry_point_names:
            entry_point_string = '{entry_point_group}:{entry_point_name}'.format(
                entry_point_group=entry_point_group, entry_point_name=inferred_entry_point_name)
        elif inferred_entry_point_name:
            entry_point_string = '{plugin_name}.{plugin_class}'.format(
                plugin_name=inferred_entry_point_name, plugin_class=plugin_class)
        else:
            # If there is no inferred entry point name, i.e. there is no module name, use an empty string as fall back
            # This should only be the case for the type string `calculation.job.JobCalculation.`
            entry_point_string = ''

        mapping_node_type_to_entry_point[type_string] = entry_point_string

    return mapping_node_type_to_entry_point



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import unittest
from tornado.ioloop import IOLoop
from tornado.gen import coroutine

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.engine.utils import exponential_backoff_retry, RefObjectStore

ITERATION = 0
MAX_ITERATIONS = 3


class TestExponentialBackoffRetry(AiidaTestCase):
    """Tests for the exponential backoff retry coroutine."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """Set up a simple authinfo and for later use."""
        super(TestExponentialBackoffRetry, cls).setUpClass(*args, **kwargs)
        cls.authinfo = orm.AuthInfo(computer=cls.computer, user=orm.User.objects.get_default())
        cls.authinfo.store()

    def test_exponential_backoff_success(self):
        """Test that exponential backoff will successfully catch exceptions as long as max_attempts is not exceeded."""
        ITERATION = 0
        loop = IOLoop()

        @coroutine
        def coro():
            """A function that will raise RuntimeError as long as ITERATION is smaller than MAX_ITERATIONS."""
            global ITERATION
            ITERATION += 1
            if ITERATION < MAX_ITERATIONS:
                raise RuntimeError

        max_attempts = MAX_ITERATIONS + 1
        loop.run_sync(lambda: exponential_backoff_retry(coro, initial_interval=0.1, max_attempts=max_attempts))

    def test_exponential_backoff_max_attempts_exceeded(self):
        """Test that exponential backoff will finally raise if max_attempts is exceeded"""
        ITERATION = 0
        loop = IOLoop()

        @coroutine
        def coro():
            """A function that will raise RuntimeError as long as ITERATION is smaller than MAX_ITERATIONS."""
            global ITERATION
            ITERATION += 1
            if ITERATION < MAX_ITERATIONS:
                raise RuntimeError

        max_attempts = MAX_ITERATIONS - 1
        with self.assertRaises(RuntimeError):
            loop.run_sync(lambda: exponential_backoff_retry(coro, initial_interval=0.1, max_attempts=max_attempts))


class RefObjectsStore(unittest.TestCase):

    def test_simple(self):
        """ Test the reference counting works """
        IDENTIFIER = 'a'
        OBJECT = 'my string'
        obj_store = RefObjectStore()

        with obj_store.get(IDENTIFIER, lambda: OBJECT) as obj:
            # Make sure we got back the same object
            self.assertIs(OBJECT, obj)

            # Now check that the reference has the correct information
            ref = obj_store._objects['a']
            self.assertEqual(OBJECT, ref._obj)
            self.assertEqual(1, ref.count)

            # Now request the object again
            with obj_store.get(IDENTIFIER) as obj2:
                # ...and check the reference has had it's count upped
                self.assertEqual(OBJECT, obj2)
                self.assertEqual(2, ref.count)

            # Now it should have been reduced
            self.assertEqual(1, ref.count)

        # Finally the store should be empty  (there are no more references)
        self.assertEqual(0, len(obj_store._objects))

    def test_get_no_constructor(self):
        """
        Test that trying to get an object that does exists and providing
        no means to construct it fails
        """
        obj_store = RefObjectStore()
        with self.assertRaises(ValueError):
            with obj_store.get('a'):
                pass

    def test_construct(self):
        """ Test that construction only gets called when used """
        IDENTIFIER = 'a'
        OBJECT = 'my string'

        # Use a list for a  single number so we can get references to it
        times_constructed = [
            0,
        ]

        def construct():
            times_constructed[0] += 1
            return OBJECT

        obj_store = RefObjectStore()
        with obj_store.get(IDENTIFIER, construct):
            self.assertEqual(1, times_constructed[0])
            with obj_store.get(IDENTIFIER, construct):
                self.assertEqual(1, times_constructed[0])

        # Now the object should be removed and so another call to get
        # should create
        with obj_store.get(IDENTIFIER, construct):
            self.assertEqual(2, times_constructed[0])



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Classes and methods for Django specific backend entities"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import typing

from aiida.backends.sqlalchemy.models.base import Base
from aiida.common.lang import type_check
from . import utils

ModelType = typing.TypeVar('ModelType')  # pylint: disable=invalid-name


class SqlaModelEntity(typing.Generic[ModelType]):
    """A mixin that adds some common SQLA backend entity methods"""

    MODEL_CLASS = None
    _dbmodel = None

    @classmethod
    def _class_check(cls):
        """Assert that the class is correctly configured"""
        assert issubclass(cls.MODEL_CLASS, Base), 'Must set the MODEL_CLASS in the derived class to a SQLA model'

    @classmethod
    def from_dbmodel(cls, dbmodel, backend):
        """
        Create a DjangoEntity from the corresponding db model class

        :param dbmodel: the model to create the entity from
        :param backend: the corresponding backend
        :return: the Django entity
        """
        from .backend import SqlaBackend  # pylint: disable=cyclic-import
        cls._class_check()
        type_check(dbmodel, cls.MODEL_CLASS)
        type_check(backend, SqlaBackend)
        entity = cls.__new__(cls)
        super(SqlaModelEntity, entity).__init__(backend)
        entity._dbmodel = utils.ModelWrapper(dbmodel)  # pylint: disable=protected-access
        return entity

    @classmethod
    def get_dbmodel_attribute_name(cls, attr_name):
        """
        Given the name of an attribute of the entity class give the corresponding name of the attribute
        in the db model.  It if doesn't exit this raises a ValueError

        :param attr_name:
        :return: the dbmodel attribute name
        :rtype: str
        """
        if hasattr(cls.MODEL_CLASS, attr_name):
            return attr_name

        raise ValueError("Unknown attribute '{}'".format(attr_name))

    def __init__(self, *args, **kwargs):
        super(SqlaModelEntity, self).__init__(*args, **kwargs)
        self._class_check()

    @property
    def dbmodel(self):
        """
        Get the underlying database model instance

        :return: the database model instance
        """
        return self._dbmodel._model  # pylint: disable=protected-access

    @property
    def id(self):  # pylint: disable=redefined-builtin, invalid-name
        """
        Get the id of this entity

        :return: the entity id
        """
        return self._dbmodel.id

    @property
    def pk(self):
        """
        Get the principal key for this entry

        :return: the principal key
        """
        return self._dbmodel.id

    @property
    def is_stored(self):
        """
        Is this entity stored?

        :return: True if stored, False otherwise
        """
        return self._dbmodel.id is not None

    def store(self):
        """
        Store this entity

        :return: the entity itself
        """
        self._dbmodel.save()
        return self



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the `PluginParamType`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.params.types.plugin import PluginParamType
from aiida.plugins.entry_point import get_entry_point_from_string


class TestPluginParamType(AiidaTestCase):
    """Tests for the `PluginParamType`."""

    def test_group_definition(self):
        """
        Test the various accepted syntaxes of defining supported entry point groups. Both single
        values as well as tuples should be allowed. The `aiida.` prefix should also be optional.
        """
        param = PluginParamType(group='calculations')
        self.assertIn('aiida.calculations', param.groups)
        self.assertTrue(len(param.groups), 1)

        param = PluginParamType(group='aiida.calculations')
        self.assertIn('aiida.calculations', param.groups)
        self.assertTrue(len(param.groups), 1)

        param = PluginParamType(group=('calculations',))
        self.assertIn('aiida.calculations', param.groups)
        self.assertTrue(len(param.groups), 1)

        param = PluginParamType(group=('aiida.calculations',))
        self.assertIn('aiida.calculations', param.groups)
        self.assertTrue(len(param.groups), 1)

        param = PluginParamType(group=('aiida.calculations', 'aiida.data'))
        self.assertIn('aiida.calculations', param.groups)
        self.assertIn('aiida.data', param.groups)
        self.assertTrue(len(param.groups), 2)

        param = PluginParamType(group=('aiida.calculations', 'data'))
        self.assertIn('aiida.calculations', param.groups)
        self.assertIn('aiida.data', param.groups)
        self.assertTrue(len(param.groups), 2)

    def test_get_entry_point_from_string(self):
        """
        Test the functionality of the get_entry_point_from_string which will take an entry point string
        and try to map it onto a valid entry point that is part of the groups defined for the parameter.
        """
        param = PluginParamType(group='transports')
        entry_point = get_entry_point_from_string('aiida.transports:ssh')

        # Invalid entry point strings
        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('aiida.transport:ssh')

        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('aiid.transports:ssh')

        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('aiida..transports:ssh')

        # Unsupported entry points for all formats
        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('aiida.data:structure')

        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('data:structure')

        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('structure')

        # Non-existent entry points for all formats
        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('aiida.transports:not_existent')

        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('transports:not_existent')

        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('not_existent')

        # Valid entry point strings
        self.assertEqual(param.get_entry_point_from_string('aiida.transports:ssh').name, entry_point.name)
        self.assertEqual(param.get_entry_point_from_string('transports:ssh').name, entry_point.name)
        self.assertEqual(param.get_entry_point_from_string('ssh').name, entry_point.name)

    def test_get_entry_point_from_ambiguous(self):
        """
        Test the functionality of the get_entry_point_from_string which will take an entry point string
        and try to map it onto a valid entry point that is part of the groups defined for the parameter.
        """
        param = PluginParamType(group=('aiida.calculations', 'aiida.parsers'))
        entry_point = get_entry_point_from_string('aiida.calculations:arithmetic.add')

        # Both groups contain entry point `arithmetic.add` so passing only name is ambiguous and should raise
        with self.assertRaises(ValueError):
            param.get_entry_point_from_string('arithmetic.add')

        # Passing PARTIAL or FULL should allow entry point to be returned
        self.assertEqual(param.get_entry_point_from_string('aiida.calculations:arithmetic.add').name, entry_point.name)
        self.assertEqual(param.get_entry_point_from_string('calculations:arithmetic.add').name, entry_point.name)

    def test_convert(self):
        """
        Test that the convert method returns the correct entry point
        """
        param = PluginParamType(group=('transports', 'data'))

        entry_point = param.convert('aiida.transports:ssh', None, None)
        self.assertEqual(entry_point.name, 'ssh')
        # self.assertTrue(isinstance(entry_point, EntryPoint))

        entry_point = param.convert('transports:ssh', None, None)
        self.assertEqual(entry_point.name, 'ssh')
        # self.assertTrue(isinstance(entry_point, EntryPoint))

        entry_point = param.convert('ssh', None, None)
        self.assertEqual(entry_point.name, 'ssh')
        # self.assertTrue(isinstance(entry_point, EntryPoint))

        entry_point = param.convert('aiida.data:structure', None, None)
        self.assertEqual(entry_point.name, 'structure')
        # self.assertTrue(isinstance(entry_point, EntryPoint))

        entry_point = param.convert('data:structure', None, None)
        self.assertEqual(entry_point.name, 'structure')
        # self.assertTrue(isinstance(entry_point, EntryPoint))

        entry_point = param.convert('structure', None, None)
        self.assertEqual(entry_point.name, 'structure')
        # self.assertTrue(isinstance(entry_point, EntryPoint))

        with self.assertRaises(click.BadParameter):
            param.convert('not_existent', None, None)

    def test_convert_load(self):
        """
        Test that the convert method returns the loaded entry point if load=True at construction time of parameter
        """
        param = PluginParamType(group=('transports', 'data'), load=True)
        entry_point_ssh = get_entry_point_from_string('aiida.transports:ssh')
        entry_point_structure = get_entry_point_from_string('aiida.data:structure')

        entry_point = param.convert('aiida.transports:ssh', None, None)
        self.assertTrue(entry_point, entry_point_ssh)

        entry_point = param.convert('transports:ssh', None, None)
        self.assertTrue(entry_point, entry_point_ssh)

        entry_point = param.convert('ssh', None, None)
        self.assertTrue(entry_point, entry_point_ssh)

        entry_point = param.convert('aiida.data:structure', None, None)
        self.assertTrue(entry_point, entry_point_structure)

        entry_point = param.convert('data:structure', None, None)
        self.assertTrue(entry_point, entry_point_structure)

        entry_point = param.convert('structure', None, None)
        self.assertTrue(entry_point, entry_point_structure)

        with self.assertRaises(click.BadParameter):
            param.convert('not_existent', None, None)

    def test_complete_single_group(self):
        """
        Test the complete method which is used for auto completion when there is only a single valid group, which
        means there should never be ambiguity and specifying a full entry point string is not necessary, however,
        when the user decides to user either a FULL or PARTIAL string anyway, the completion should match that syntax
        """
        param = PluginParamType(group=('transports'))
        entry_point_minimal = 'ssh'
        entry_point_partial = 'transports:ssh'
        entry_point_full = 'aiida.transports:ssh'

        options = [item[0] for item in param.complete(None, 'ss')]
        self.assertIn(entry_point_minimal, options)

        options = [item[0] for item in param.complete(None, 'ssh')]
        self.assertIn(entry_point_minimal, options)

        options = [item[0] for item in param.complete(None, 'transports:ss')]
        self.assertIn(entry_point_partial, options)

        options = [item[0] for item in param.complete(None, 'transports:ssh')]
        self.assertIn(entry_point_partial, options)

        options = [item[0] for item in param.complete(None, 'aiida.transports:ss')]
        self.assertIn(entry_point_full, options)

        options = [item[0] for item in param.complete(None, 'aiida.transports:ssh')]
        self.assertIn(entry_point_full, options)

    def test_complete_amibguity(self):
        """
        Test the complete method which is used for auto completion when the supported groups share an entry point
        with the same name, which can lead to ambiguity. In this case the autocomplete should always return the
        possibilites in the FULL entry point string format. When the user tries to autocomplete
        """
        param = PluginParamType(group=('aiida.calculations', 'aiida.parsers'))
        entry_point_full_calculations = 'aiida.calculations:arithmetic.add'
        entry_point_full_parsers = 'aiida.parsers:arithmetic.add'

        options = [item[0] for item in param.complete(None, 'aiida.calculations:arith')]
        self.assertIn(entry_point_full_calculations, options)

        options = [item[0] for item in param.complete(None, 'aiida.calculations:arithmetic.add')]
        self.assertIn(entry_point_full_calculations, options)

        options = [item[0] for item in param.complete(None, 'aiida.parsers:arith')]
        self.assertIn(entry_point_full_parsers, options)

        options = [item[0] for item in param.complete(None, 'aiida.parsers:arithmetic.add')]
        self.assertIn(entry_point_full_parsers, options)

        # PARTIAL or MINIMAL string formats will not be autocompleted
        options = [item[0] for item in param.complete(None, 'parsers:arith')]
        self.assertNotIn(entry_point_full_calculations, options)
        self.assertNotIn(entry_point_full_parsers, options)

        options = [item[0] for item in param.complete(None, 'parsers:arithmetic.add')]
        self.assertNotIn(entry_point_full_calculations, options)
        self.assertNotIn(entry_point_full_parsers, options)

        options = [item[0] for item in param.complete(None, 'arith')]
        self.assertNotIn(entry_point_full_calculations, options)
        self.assertNotIn(entry_point_full_parsers, options)

        options = [item[0] for item in param.complete(None, 'arithmetic.add')]
        self.assertNotIn(entry_point_full_calculations, options)
        self.assertNotIn(entry_point_full_parsers, options)
